/**************************************************************************
 *
 *  This file is part of the Coin SoWin GUI binding library.
 *  Copyright (C) 2000 by Systems in Motion.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 as published by the Free Software Foundation.  See the file
 *  LICENSE.LGPL at the root directory of the distribution for all the
 *  details.
 *
 *  If you want to use Coin SoWin for applications not compatible with the
 *  LGPL, please contact SIM to acquire a Professional Edition License.
 *
 *  Systems in Motion, Prof Brochs gate 6, N-7030 Trondheim, NORWAY
 *  http://www.sim.no/ support@sim.no Voice: +47 22114160 Fax: +47 22207097
 *
 **************************************************************************/

// 

#ifndef SOWIN_MATERIALLIST_H
#define SOWIN_MATERIALLIST_H

#include <Inventor/Win/SoWinComponent.h>

class SoMaterial;
class SoAnyMaterialList;

typedef void SoWinMaterialListCB(void * closure, const SoMaterial * material);

// *************************************************************************

class SOWIN_DLL_API SoWinMaterialList : public SoWinComponent {
  SOWIN_OBJECT_HEADER(SoWinMaterialList, SoWinComponent);

public:
  SoWinMaterialList(
    HWND parent = NULL,
    const char * const name = NULL,
    SbBool embed = TRUE,
    const char * const dir = NULL);
  ~SoWinMaterialList(void);

  void addCallback(
    SoWinMaterialListCB * const callback,
    void * const closure = NULL);
  void removeCallback(
    SoWinMaterialListCB * const callback,
    void * const closure = NULL);

protected:
  SoWinMaterialList(
    HWND parent,
    const char * const name,
    SbBool embed,
    const char * const dir,
    SbBool build);

  virtual const char * getDefaultWidgetName(void) const;
  virtual const char * getDefaultTitle(void) const;
  virtual const char * getDefaultIconTitle(void) const;

  HWND buildWidget(HWND parent);
  HWND buildPulldownMenu(HWND parent);

private:
  void constructor(const char * const dir, const SbBool build);

  // FIXME: remove all this and set it up for use with a
  // SoGuiMaterialListP class. 2001???? larsa.

  void selectionCallback(int i);
  static void selection_cb(HWND, void *, void *);

  void paletteMenuCallback(HWND);
  static void palette_menu_cb(HWND, void *, void *);


#ifdef __COIN_SOGTK__
  void menuactivation(GtkWidget * menuitem);
  static void menuactivationCB(GtkObject * obj, gpointer closure);

  void itemactivation(int item);
  static void itemactivationCB(GtkObject * obj, gint row, gint column,
                               GdkEvent * event, gpointer closure);
#endif // __COIN_SOGTK__

  HWND listwidget;

  SoAnyMaterialList * common;

}; // public SoWinMaterialList

// *************************************************************************

#endif // ! SOWIN_MATERIALLIST_H
