
#include <Inventor/Win/SoWin.h>
#include <Inventor/Win/viewers/SoWinExaminerViewer.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoCube.h>

#include <windowsx.h>

#define Widget HWND
#define SIZE 400
#define ID_EXIT 314

static LRESULT CALLBACK WindowProc (HWND,UINT,WPARAM,LPARAM);
static const char className[] = "MyWindow";

class So_WinExaminerViewer : public SoWinExaminerViewer {
public:
  So_WinExaminerViewer(HWND parent,const char* name,SbBool embed)
  :  SoWinExaminerViewer(parent,name,embed){}
  virtual void sizeChanged(const SbVec2s & size){
    SoWinExaminerViewer::sizeChanged(size);
  }
};

//////////////////////////////////////////////////////////////////////////////
int main(
 int
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Widget top = SoWin::init(aArgv[0]);
  if (top==NULL) exit(1);

  // Create a scene :

  SoSeparator* root=new SoSeparator;
  root->ref();

  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue (SbColor(1,1,0));
  root->addChild(material);

  SoScale* scale = new SoScale;
  scale->scaleFactor.setValue(0.1F,0.1F,0.1F);
  //scale->scaleFactor.setValue(0.5F,0.5F,0.5F);
  root->addChild(scale);

  SoCube* cube = new SoCube;
  root->addChild(cube);

  //Create a shell window :
  static SbBool done = FALSE;
  if(done==FALSE) {
    HBRUSH brush = (HBRUSH) GetSysColorBrush(COLOR_BTNFACE);
    WNDCLASS wc;
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC)WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = ::GetModuleHandle(0);
    wc.hIcon = ::LoadIcon(0, IDI_APPLICATION);
    wc.hCursor = ::LoadCursor(0, IDC_ARROW);
    wc.hbrBackground = brush;
    wc.lpszMenuName = className;
    wc.lpszClassName = className;
    ::RegisterClass(&wc);
    done = TRUE;
  }


  HMENU menuBar = CreateMenu();
  HMENU casc = CreatePopupMenu();
  ::AppendMenu(menuBar,MF_POPUP,(UINT)casc,"File");
  ::AppendMenu(casc,MF_STRING,ID_EXIT,"Exit");

  //  Compell window to be created at 0,0 to bypass 
  // the 'black border' problem. 
  HWND parent = ::CreateWindow(className, "myWindow", 
                                   WS_OVERLAPPEDWINDOW |
                                   WS_VISIBLE |
                                   WS_CLIPSIBLINGS |
                                   WS_CLIPCHILDREN,
                                   CW_USEDEFAULT, CW_USEDEFAULT, 
                                   //0,0,
                                   SIZE,SIZE,
                                   0, menuBar,::GetModuleHandle(0),0);
  // Retreive window and client sizez :
  RECT wrect,crect;
  GetWindowRect((HWND)parent,&wrect);
  GetClientRect((HWND)parent,&crect);
  int ww = wrect.right-wrect.left;
  int wh = wrect.bottom-wrect.top;
  int cw = crect.right-crect.left;
  int ch = crect.bottom-crect.top;
  // Compell client rect to be SIZE SIZE :
  MoveWindow((HWND)parent,wrect.left,wrect.top,SIZE+ww-cw,SIZE+wh-ch,TRUE);
  ::SetWindowText((HWND)parent,"myWindow");

  SoWinExaminerViewer* viewer = new So_WinExaminerViewer(parent,"xxx",TRUE);
  ::SetWindowLong((HWND)parent,GWL_USERDATA,LONG(viewer));
  viewer->setSize(SbVec2s(SIZE,SIZE));
  viewer->setBackgroundColor(SbColor(0.4F,0.4F,0.4F));
  viewer->setSceneGraph(root);
  viewer->viewAll();
  //viewer->show();
  viewer->setTitle("CoinWin");
  //viewer->setViewing(FALSE);
  //viewer->setDecoration(FALSE);
  //viewer->setAutoClipping(FALSE);

  //SoWin::show(top);
  SoWin::show(parent);

  SoWin::mainLoop();

  //Cleanup
  root->unref();

  delete viewer;

  //SoWin::clear();  

  return 0;
}
//////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK WindowProc ( 
 HWND   aWindow
,UINT   aMessage
,WPARAM aWParam
,LPARAM aLParam
)
//////////////////////////////////////////////////////////////////////////////
//  Below treatment of WM_SIZE, WM_SETFOCUS not necessary 
// with TGS, but needed with SoFree. WM_DESTROY needed for 
// 'main top level window' so that 'Close window' induces
// the end of the task.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  switch (aMessage) { 
  case WM_SIZE:{ // Assume one child window !
    int width = LOWORD(aLParam);
    int height = HIWORD(aLParam);
    //printf("debug : G4SoWindow : WMS_SIZE : %d %d\n",width,height);
    So_WinExaminerViewer* viewer = 
      (So_WinExaminerViewer*)::GetWindowLong(aWindow,GWL_USERDATA);
    if(viewer) viewer->sizeChanged(SbVec2s(width, height));
  }return 0;
  case WM_SETFOCUS:{ // Assume one child window !
    HWND hwnd = ::GetFirstChild(aWindow);
    if(hwnd!=0) ::SetFocus(hwnd);
  }return 0;
  case WM_DESTROY:{
    //G4OpenInventorWinViewer* This = 
    //  (G4OpenInventorWinViewer*)::GetWindowLong(aWindow,GWL_USERDATA);
    //::PostQuitMessage(0);
  }return 0;
  case WM_COMMAND:{
    if(aWParam==ID_EXIT) ::PostQuitMessage(0);
  }return 0;
  default:
    return (::DefWindowProc(aWindow,aMessage,aWParam,aLParam));
  }
}
