#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoFaceSet.h>

#include <HEPVis/SbColors.h>
#include <HEPVis/misc/SoStyleCache.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/draggers/SoDragPoint2D.h>

#include <stdlib.h> //exit

static void finish_CB(void*,SoDragger* aDragger) {
  SoDragPoint2D* dragger = (SoDragPoint2D*)aDragger;
  const SbVec3f& v = dragger->point();
  printf("debug : finish_CB : %g %g %g\n",v[0],v[1],v[2]);
}

int main(int aArgc,char** aArgv){

  // Set up the main window
  SoGuiWidget topWidget=SoGui::init(aArgv[0]);
  if (!topWidget) exit(1);

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoStyleCache* styleCache = new SoStyleCache;
  styleCache->ref();

  int ptn = 6;
  SbVec3f pts[6];
  pts[0].setValue(0,0,0);
  pts[1].setValue(1,0,0);
  pts[2].setValue(1.5,0.5,0);
  pts[3].setValue(0.5,1.5,0);
  pts[4].setValue(0,1,0);
  pts[ptn-1] = pts[0]; //close the polyline.

 {SoSeparator* separator = new SoSeparator;
  sceneGraph->addChild(separator);

  separator->addChild(styleCache->getFilled());
  separator->addChild(styleCache->getNormalBindingOverall());
  separator->addChild(styleCache->getNormalZ());
  separator->addChild(styleCache->getMaterial(SbColor_grey,0.5));

  SoCoordinate3* coordinate3 = new SoCoordinate3;
  coordinate3->point.setValues(0,ptn,pts);
  separator->addChild(coordinate3);

  SoFaceSet* faceSet = new SoFaceSet;
  faceSet->numVertices.setValues(0,1,&ptn);
  separator->addChild(faceSet);

  separator->addChild(styleCache->getLineStyle(SbLinePattern_solid,1));
  separator->addChild(styleCache->getMaterial(SbColor_red,0));

  SoLineSet* lineSet = new SoLineSet;
  lineSet->numVertices.setValues(0,1,&ptn);
  separator->addChild(lineSet); 

 {SoSeparator* separator = new SoSeparator;
  sceneGraph->addChild(separator);
  SoTools::addDraggersToPolygon(*separator,*coordinate3,
                                SbColor_green,0.08F,finish_CB,0);}

  } // end add in scene graph.

  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(topWidget);
  examinerViewer->setSceneGraph(sceneGraph);
  examinerViewer->show();

  SoGui::show(topWidget);
  SoGui::mainLoop();

  styleCache->unref();

  sceneGraph->unref();
  delete examinerViewer;

  return 0; 
}
