#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLineSet.h>

#include <HEPVis/SbArguments.h>
#include <HEPVis/misc/SoTools.h> //For SoTools::write.
#include <HEPVis/nodes/SoHatching.h>
#include <HEPVis/nodes/SoTextHershey.h>
#include <HEPVis/nodes/SoTextTTF.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

#include <stdlib.h> //rand

#include <iostream>

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SbArguments args(aArgc,aArgv);
  if(args.isAnArgument("-h")) {
    std::cout << "HEPVis::HatchBox :" << std::endl;
    std::cout << "Usage :" << std::endl;
    std::cout << " OS> HatchBox<GUI> <options>" << std::endl;
    std::cout << "Options :" << std::endl;
    std::cout << " -h : to get help." << std::endl;
    std::cout << " -altrep : to produce an alternate representation." 
              << std::endl;
    std::cout << " -write : to write scene in a file (default is out.iv)." 
              << std::endl;
    std::cout << " -file : name of the file (with .iv or .wrl extension)." 
              << std::endl;
    std::cout << "For example :" << std::endl;
    std::cout << " OS> HatchBox -altrep -write" << std::endl;
    std::cout << " UNIX> ivview out.iv" << std::endl;
    return 0;
  }

  // Set up the main window
  SoGuiWidget topWidget=SoGui::init(aArgv[0]);
  if (!topWidget) exit(1);
  
  // initialize HEPVis used classes :
  SoHatching::initClass();
  SoTextHershey::initClass();
  SoTextTTF::initClass();
  SoAlternateRepAction::initClass(); //For SoTools::write.
  
  SoSeparator* root=new SoSeparator;
  root->ref();
  
  SoLightModel* lightModel = new SoLightModel;
  lightModel->model.setValue(SoLightModel::BASE_COLOR);
  root->addChild(lightModel);

  SbVec3f* list4 = new SbVec3f[55];
  SoHatching* hatch7 = new SoHatching;
  
  list4[0].setValue(9.000000,2.000000,2.000000);
  list4[1].setValue(10.000000,6.000000,8.000000);
  list4[2].setValue(9.000000,4.000000,9.000000);
  list4[3].setValue(29.000000,120.000000,255.000000);
  list4[4].setValue(25.000000,114.000000,266.000000);
  list4[5].setValue(53.000000,210.000000,378.000000);
  list4[6].setValue(13.000000,70.000000,208.000000);
  list4[7].setValue(39.000000,144.000000,259.000000);
  list4[8].setValue(40.000000,142.000000,244.000000);
  list4[9].setValue(45.000000,160.000000,267.000000);
  list4[10].setValue(43.000000,224.000000,507.000000);
  list4[11].setValue(31.000000,94.000000,148.000000);
  list4[12].setValue(13.000000,46.000000,124.000000);
  list4[13].setValue(56.000000,230.000000,424.000000);
  list4[14].setValue(21.000000,114.000000,298.000000);
  list4[15].setValue(46.000000,228.000000,497.000000);
  list4[16].setValue(11.000000,32.000000,91.000000);
  list4[17].setValue(31.000000,130.000000,274.000000);
  list4[18].setValue(44.000000,148.000000,233.000000);
  list4[19].setValue(18.000000,138.000000,406.000000);
  list4[20].setValue(56.000000,202.000000,326.000000);
  list4[21].setValue(16.000000,52.000000,121.000000);
  list4[22].setValue(18.000000,98.000000,266.000000);
  list4[23].setValue(43.000000,156.000000,269.000000);
  list4[24].setValue(48.000000,228.000000,481.000000);
  list4[25].setValue(43.000000,202.000000,430.000000);
  list4[26].setValue(14.000000,36.000000,81.000000);
  list4[27].setValue(11.000000,66.000000,210.000000);
  list4[28].setValue(26.000000,134.000000,328.000000);
  list4[29].setValue(42.000000,196.000000,417.000000);
  list4[30].setValue(47.000000,208.000000,419.000000);
  list4[31].setValue(57.000000,234.000000,430.000000);
  list4[32].setValue(29.000000,128.000000,283.000000);
  list4[33].setValue(28.000000,150.000000,368.000000);
  list4[34].setValue(17.000000,64.000000,155.000000);
  list4[35].setValue(26.000000,124.000000,293.000000);
  list4[36].setValue(37.000000,118.000000,184.000000);
  list4[37].setValue(60.000000,272.000000,539.000000);
  list4[38].setValue(40.000000,178.000000,370.000000);
  list4[39].setValue(59.000000,262.000000,512.000000);
  list4[40].setValue(54.000000,240.000000,475.000000);
  list4[41].setValue(30.000000,156.000000,373.000000);
  list4[42].setValue(19.000000,132.000000,377.000000);
  list4[43].setValue(38.000000,128.000000,211.000000);
  list4[44].setValue(21.000000,96.000000,235.000000);
  list4[45].setValue(31.000000,158.000000,372.000000);
  list4[46].setValue(39.000000,162.000000,322.000000);
  list4[47].setValue(46.000000,164.000000,273.000000);
  list4[48].setValue(26.000000,140.000000,349.000000);
  list4[49].setValue(37.000000,122.000000,198.000000);
  list4[50].setValue(36.000000,144.000000,283.000000);
  list4[51].setValue(18.000000,92.000000,245.000000);
  list4[52].setValue(54.000000,248.000000,503.000000);
  list4[53].setValue(25.000000,126.000000,308.000000);
  list4[54].setValue(9.000000,2.000000,2.000000);
  hatch7->spacing.setValue(-1.66F);
  hatch7->offset.setValue(0.2F);

  if (hatch7->addPolyline(list4,55) == TRUE) {
    root->addChild(hatch7); 
      
    SoCoordinate3* coordinate3 = new SoCoordinate3;
    int32_t pointn = 55;
    coordinate3->point.setValues(0,pointn,list4);
    root->addChild(coordinate3);

    SoLineSet* lineSet = new SoLineSet;
    lineSet->numVertices.set1Value(0,55);
    root->addChild(lineSet);

  } else {
    hatch7->unref();
  }

  printf("End of computing...\n");
  delete [] list4;

  if (args.isAnArgument("-write")) {
    SbBool altRep = args.isAnArgument("-altrep")?TRUE:FALSE;
    SbString file;
    args.find("-file",file);
    if(file=="") file = "out.iv";
    SoTools::write(*root,file,"",FALSE,altRep);
  }

  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(topWidget);
  examinerViewer->setSceneGraph(root);
  examinerViewer->viewAll();
  examinerViewer->show();

  SoGui::show(topWidget);
  SoGui::mainLoop();

  root->unref();
  delete examinerViewer;
#ifdef SoFreePackage //Cleanup
  SoGui::clear();  
#endif

  return 0; 
}
