#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/nodes/SoHelicalTrack.h>

// Test of SoHelicalTrack class
// Joe Boudreau Nov. 7, 1996
//
// This demo draws a handfull of tracks:

#define NTRACKS 24
float TrackParameters[NTRACKS][5]= {
  {0.0235598F  ,-2.64028F  ,0.223126F,-0.000947721F,-19.2387F},
  {-0.026806F  ,-1.85632F  ,3.03883F ,0.000792523F,-19.2367F},
  {0.0226147F  ,0.517672F  ,1.25412F ,-0.000769267F,-19.2382F},
  {-0.0102705F ,-1.00757F  ,1.10671F ,-0.0008555F,-19.2373F},
  {-0.00616535F,-1.37573F  ,1.55067F ,-0.000547534F,-19.2367F},
  {0.0106418F  ,-2.46822F  ,4.55818F ,0.000654988F,-19.2397F},
  {0.0106948F  ,0.751094F  ,2.13166F ,3.67214e-06F,-19.2386F},
  {-0.00473071F,-2.86888F  ,4.96651F ,0.000299413F,-19.2406F},
  {0.0430075F  ,-0.64392F  ,3.89761F ,0.000983425F,-19.238F},
  {-0.00814083F,-1.27395F  ,0.774761F,-0.000979539F,-19.2376F},
  {0.00299336F ,1.63521F   ,3.50505F ,0.000984415F,-19.2382F},
  {-0.0419553F ,3.24183F   ,0.864165F,-0.000956295F,-19.2389F},
  {0.0149561F  ,-0.715608F ,4.94913F ,0.000316014F,-19.2385F},
  {-0.0125898F ,0.188497F  ,3.8315F  ,0.000994365F,-19.2378F},
  {-0.0150922F ,-0.662462F ,2.65618F ,0.000505558F,-19.2373F},
  {-0.0113685F ,-0.813852F ,1.1756F  ,-0.000817407F,-19.2374F},
  {0.00974994F ,-2.9294F   ,5.43874F ,-0.000168868F,-19.2408F},
  {0.00899237F ,-3.17087F  ,2.86856F ,0.00067685F,-19.2355F},
  {0.00591087F ,-0.0254306F,2.32928F ,0.000200557F,-19.2378F},
  {-0.0113543F ,-1.48355F  ,2.13656F ,8.57914e-06F,-19.2364F},
  {0.0156185F  ,-3.58749F  ,5.73841F ,-0.00045326F,-19.2411F},
  {-0.00936308F,-2.23641F  ,3.05724F ,0.000803715F,-19.2365F},
  {0.0218428F  ,-0.0757606F,0.365364F,0.916965F,-22.1299F},
  {-0.0358958F ,-2.97726F  ,4.89028F ,-1.53537F,-20.3828F}
};


int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoHelicalTrack::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  // Set up the light model and the material for
  // The object to be drawn

  SoMaterial    *TrackMaterial = new SoMaterial;
  TrackMaterial->diffuseColor.setValue(1,0,0);
  root->addChild(TrackMaterial);

  SoLightModel  *TrackModel    = new SoLightModel;
  TrackModel->model = SoLightModel::BASE_COLOR;
  root->addChild(TrackModel);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  SoTimeCounter *TheCounter = new SoTimeCounter;
  TheCounter->max = 100; 
  TheCounter->min = 1;
  TheCounter->frequency=0.2F;

  // Make the Helical Tracks

  for (int i=0;i<NTRACKS;i++) {
    SoHelicalTrack      *TheHelicalTrack   = new SoHelicalTrack;
    TheHelicalTrack->inverseRadius.setValue(TrackParameters[i][0]);
    TheHelicalTrack->cotTheta.setValue(TrackParameters[i][1]);
    TheHelicalTrack->phi0.setValue(TrackParameters[i][2]);
    TheHelicalTrack->d0.setValue(TrackParameters[i][3]);
    TheHelicalTrack->z0.setValue(TrackParameters[i][4]);

    // Generate the alternate representation to generate an inventor
    // file readable by anyone, and write it out. This is an important
    // feature of any new HepVis node.

    TheHelicalTrack->s1.setValue(50);
    if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP"))TheHelicalTrack->generateAlternateRep(0);
    TheHelicalTrack->s1.connectFrom(&TheCounter->output);
    root->addChild(TheHelicalTrack);
  }

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer

  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->show();

  SoGui::show(myWindow);
  myViewer->getCamera()->position.setValue(SbVec3f(0,0,100));
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoHelicalTrack::clearClass();
  SoGui::clear();  
#endif

  return 0; 
}

