#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/SoDB.h>
#include <Inventor/fields/SoField.h>

#include <HEPVis/nodekits/SoIdealTrackKit.h>
#include <HEPVis/nodekits/SoIdealBeamTrackKit.h>

//
// Test of SoIdealTrackKit class, after JB
//
// This demo draws a IdealTrackKit
//
 
int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class
  ParticleChar::initParticles(); //init particle database
  SoIdealTrackKit::initClass();
  SoIdealBeamTrackKit::initClass();


  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();


  // I like to use a global field here and then connect it to all the tracks.
  // This could be local, but global fits better to the overall concept of
  // time...
  SoField *timeDeltaField;
  if  ((timeDeltaField = SoDB::getGlobalField("timeDeltaField")) == NULL){
    timeDeltaField = SoDB::createGlobalField("timeDeltaField",
                     SoSFFloat::getClassTypeId());
    ((SoSFFloat *)timeDeltaField)->setValue((float) 500);
  }
    
  SoTimeCounter *TheCounter = new SoTimeCounter;
  TheCounter->max  = 400;
  TheCounter->min = -200; 
  TheCounter->frequency = 0.2F;
  
  // Find the electron characteristics...
  ParticleChar *electron = ParticleChar::getByName("electron");
  // Make the IdealTrackKit  
  SoIdealTrackKit *TheIdealTrackKit   = new SoIdealTrackKit();
  // Set the kinematics
  TheIdealTrackKit->initTrack(0.,0.,0.,0.0,0.5,0.2,0.0,electron);
  root->addChild(TheIdealTrackKit);

  // Beam tracks need to be able to go backwards in time from the vertex.
/*
  SoIdealBeamTrackKit *beamPlus = new SoIdealBeamTrackKit;
  ParticleChar *proton = ParticleChar::getByName("proton");
  beamPlus->initTrack(0.,0.,0.,0,0,7000,0,proton);
  root->addChild(beamPlus);
  SoIdealBeamTrackKit *beamMinus = new SoIdealBeamTrackKit;
  beamMinus->initTrack(0.,0.,0.,0,0,-7000,0,proton);
  root->addChild(beamMinus);

  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP"))  {
          TheIdealTrackKit->generateAlternateRep(0);
        beamPlus->generateAlternateRep(0);
        beamMinus->generateAlternateRep(0);
        }
*/  

/*
  TheIdealTrackKit->DeltaT.connectFrom(timeDeltaField);
  beamPlus->DeltaT.connectFrom(timeDeltaField);
  beamMinus->DeltaT.connectFrom(timeDeltaField);
  timeDeltaField->connectFrom(&TheCounter->output);
*/  
  //debug TheIdealTrackKit->DeltaT.connectFrom(timeDeltaField);
//  beamPlus->DeltaT.connectFrom(timeDeltaField);
//  beamMinus->DeltaT.connectFrom(timeDeltaField);
    ((SoSFFloat *)timeDeltaField)->setValue((float) 5000.0);
  
  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer
  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->viewAll();
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

#ifdef SoFreePackage //Cleanup
  root->unref();
  delete WriteAction;
  delete myViewer;
  SoIdealTrackKit::clearClass();
  SoGui::clear();  
  printf("IdealTrackKit : exiting.\n");
#endif

  return 0; 
}

