#include <HEPVis/Gui/SoGui.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/errors/SoDebugError.h>

// HEPVis :
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodes/SoImage.h>
#include <HEPVis/nodekits/SoImageRegion.h>
#include <HEPVis/nodekits/SoCanvas.h>
#include <HEPVis/actions/SoGL2PSAction.h>

extern "C" {
  void HEPVisUtils_initClasses();
}

#include <math.h>
#include <stdlib.h>
#include <string.h>

void keystroke(void*,SoEventCallback*);

SoGuiExaminerViewer* examinerViewer = NULL;

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  HEPVis_SoImage::Format fmt = HEPVis_SoImage::GIF;
  if(aArgc==2) {
    // A file name is given as first argument.
    // Guess its type (GIF ir FITS).
    if(strstr(aArgs[1],".gif")!=NULL) 
      fmt = HEPVis_SoImage::GIF;
    else if(strstr(aArgs[1],".jpg")!=NULL) 
      fmt = HEPVis_SoImage::JPEG;
    else if(strstr(aArgs[1],".fits")!=NULL) 
      fmt = HEPVis_SoImage::FITS;
    else if(strstr(aArgs[1],".fit")!=NULL) 
      fmt = HEPVis_SoImage::FITS;
    else if(strstr(aArgs[1],".tex")!=NULL) 
      fmt = HEPVis_SoImage::TEX;
    else if(strstr(aArgs[1],".dvi")!=NULL) 
      fmt = HEPVis_SoImage::DVI;
    else {
      printf("File extension .gif,.jpg,.fits,.fit,.tex,.dvi expected for \"%s\".\n",aArgs[1]);
      return 0;
    }
  } else {
    printf("Give a .gif or a .fit file as first argument.\n");
    printf("The HEPVis/data directory contains some.\n");
    return 0;
  }
    

  SoGuiWidget myWindow = SoGui::init(aArgs[0]);
  if (myWindow==NULL) exit(1);

  HEPVisUtils_initClasses();

  SoCanvas* canvas = new SoCanvas;
  canvas->ref();

  SoSeparator* sceneGraph = canvas->getTopSeparator();

  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),keystroke,0);
  sceneGraph->addChild(eventCallback);                                  

  SoImageRegion* imageRegion = new SoImageRegion;
  SoViewportRegion* viewportRegion = imageRegion->getViewportRegion();
  viewportRegion->setPositionPercent(0.1F,0.25F);
  viewportRegion->setSizePercent(0.8F,0.5F);
  viewportRegion->horizontalShadow.setValue(10);
  viewportRegion->verticalShadow.setValue(-10);
  imageRegion->horizontalMargin.setValue(20);
  imageRegion->verticalMargin.setValue(20);
  HEPVis_SoImage* pixmap = imageRegion->getImage();
  pixmap->fileName.setValue(aArgs[1]);
  pixmap->format.setValue(fmt);
  canvas->getTopSeparator()->addChild(imageRegion);

  // View the new object in an Examiner Viewer

  examinerViewer = new SoGuiExaminerViewer(myWindow);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSize(SbVec2s(400,400));
  examinerViewer->setSceneGraph(canvas);
  examinerViewer->viewAll();
  examinerViewer->show();
  examinerViewer->setBackgroundColor(SbColor(0.6F,0.6F,0.6F));

  SoDebugError::postInfo("Image",
    "type P key in picking mode to produce a postscript file.");

  SoGui::show(myWindow);
  SoGui::mainLoop();

  canvas->unref();
  delete examinerViewer;
  delete action;

  return EXIT_SUCCESS;
}
//////////////////////////////////////////////////////////////////////////////
void keystroke( 
 void*
,SoEventCallback* eventCB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(event,P)) {
    if(examinerViewer) {
      SoDebugError::postInfo("Image","produce out.ps...");

      SoGL2PSAction* action = 
        (SoGL2PSAction*)examinerViewer->getGLRenderAction();
      action->enableFileWriting();
      examinerViewer->render();
      action->disableFileWriting();

      eventCB->setHandled();
    }
  }
}

