#include <HEPVis/Gui/SoGui.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>

// HEPVisUtils :
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/actions/SoGL2PSAction.h>

// HEPVisPlotter :
#include <HEPVis/nodekits/SoInfosRegion.h>

extern "C" {
  void HEPVisUtils_initClasses();
  void HEPVisPlotter_initClasses();
}

// Else :
#include <stdlib.h>
#include <stdio.h>

static void keystroke(void*,SoEventCallback*);

static SoGuiExaminerViewer* examinerViewer = NULL;

//#include "dsm.h"

int main(int aArgc,char** aArgv){
  //LookDSM_Problem(); //G.Barrand

  SoGuiWidget widget=SoGui::init(aArgv[0]);
  if (widget==NULL) exit(1);

  HEPVisUtils_initClasses();
  HEPVisPlotter_initClasses();

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoInfosRegion* infosRegion = new SoInfosRegion;
  infosRegion->text.set1Value(0,"item_0");
  infosRegion->text.set1Value(1,"value_0");
  infosRegion->text.set1Value(2,"item_1");
  infosRegion->text.set1Value(3,"value_1");
  SoViewportRegion* viewportRegion = infosRegion->getViewportRegion();
  viewportRegion->setSizePercent(0.75F,0.7F);
  viewportRegion->backgroundColor.setValue(SbColor(0.5F,0.5F,0.5F));
  SoTextStyle* soStyle = infosRegion->getTextStyle();
  soStyle->fontName.setValue("Hershey");

  sceneGraph->addChild(infosRegion);

  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                  keystroke,0);
  sceneGraph->addChild(eventCallback);                                  

  ////////////////////////////////////////////////////////////

  examinerViewer = new SoGuiExaminerViewer(widget);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSize(SbVec2s(700,450));
  examinerViewer->setAutoClipping(FALSE);
  examinerViewer->setViewing(FALSE);
  examinerViewer->setHeadlight(FALSE);
  examinerViewer->setDecoration(FALSE);
  examinerViewer->setBackgroundColor(SbColor(0.5F,0.5F,0.5F));
  examinerViewer->setSceneGraph(sceneGraph);
  examinerViewer->show();

  SoGui::show(widget);

  SoDebugError::postInfo("Plotter","P key in picking mode : produce vector PS (out.ps) of the current page.");

  SoGui::mainLoop();

  //Cleanup
  sceneGraph->unref();
  delete examinerViewer;

  return EXIT_SUCCESS;
}
//////////////////////////////////////////////////////////////////////////////
void keystroke( 
 void*
,SoEventCallback* eventCB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(event,P)) {
    if(examinerViewer) {
      SoDebugError::postInfo("Plotter","produce out.ps...");

      SoGL2PSAction* action = 
        (SoGL2PSAction*)examinerViewer->getGLRenderAction();
      action->enableFileWriting();
      printf("debug : GL2PS : render : 000\n");
      examinerViewer->render();
      printf("debug : GL2PS : render : 001\n");
      action->disableFileWriting();

      eventCB->setHandled();
    }
  }
}
