////////////////////////////////////////////////////////////
//
// Created by Laurent Garnier to demonstrate and test the
// visualization of a Latex with the HEPVis_SoImage node.
//
////////////////////////////////////////////////////////////

#include <HEPVis/Gui/SoGui.h>

// Inventor :

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/errors/SoDebugError.h>

// HEPVis :
#include <HEPVis/SbLatex.h>
#include <HEPVis/SbDVI.h>
#include <HEPVis/nodes/SoImage.h>
#include <HEPVis/actions/SoGL2PSAction.h>

#include <math.h>
#include <stdlib.h>

void keystroke(void*,SoEventCallback*);
SbString  ex1();
SbString  ex2();
SbString  ex3();

SoGuiExaminerViewer* examinerViewer = NULL;

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  /* here is the code to get a file from the input or a function above
     AND modify parameters in dvi process

  SbString dviTmpName = "";
  unsigned int width;
  unsigned int height;
  unsigned char* gifData;
  int colorNumber;
  double* reds;
  double* greens;
  double* blues;
  int transparent;

  SbDVI dvi;
  SbLatex latex;
  if (aArgc ==1) {
    //
    SbString texstring = "\\documentclass[11pt]{article} \n";
    texstring += "\\usepackage{color}\n";
    texstring += "\\pagestyle{empty}\n";
    texstring += "\\begin{document}\n";
    texstring += ex3();
    texstring += "\\end{document}  \n";
    
    if (!latex.strToDvi(texstring,dviTmpName))
      exit(1);
  } else if (aArgc ==2) {
    if (!latex.fileToDvi(aArgs[1],dviTmpName))
      exit(1);
  }
  //  dvi.setFontGen(true);
  dvi.setTransparent(false);
  //  dvi.setBlur(true);
  //  dvi.setBackground("#0000FF");
  //  dvi.setForeground("#FF0000");
  //  dvi.setResolution(110);

    if (!dvi.dviToGIF(dviTmpName.getString(),gifData,width,height,colorNumber,
                      reds,greens,blues,transparent))
      exit(1);
  // remove dvi tmp file
  remove(dviTmpName.getString());
  */

  SoGuiWidget myWindow = SoGui::init(aArgs[0]);
  if (myWindow==NULL) exit(1);

  HEPVis_SoImage::initClass();
  SoGL2PSAction::initClass();

  SoSeparator* root = new SoSeparator;
  root->ref();

  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(1.0, 1.0, 0.0);
  root->addChild(material);

  SoCube* cube = new SoCube;
  root->addChild(cube);

  SoTransform* transform = new SoTransform;
  transform->translation.setValue(1.1F,1.1F,0);
  root->addChild(transform);

  HEPVis_SoImage* pixmap = new HEPVis_SoImage;

  pixmap->fileName.setValue(aArgc==2?aArgs[1]:"");
  pixmap->format.setValue(HEPVis_SoImage::TEX);

  // here is the code to get a file from the input or a function above
  //   AND modify parameters in dvi process
  //  pixmap->setPixmap(width,height,gifData,
  //                    colorNumber,reds,greens,blues,transparent);

  root->addChild(pixmap);

  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                  keystroke,0);
  root->addChild(eventCallback);                                  

  // View the new object in an Examiner Viewer

  examinerViewer = new SoGuiExaminerViewer(myWindow);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSize(SbVec2s(400,400));
  examinerViewer->setSceneGraph(root);
  examinerViewer->viewAll();
  examinerViewer->show();
  examinerViewer->setBackgroundColor(SbColor(0.0,0.0,0.0));

  SoDebugError::postInfo("Image",
    "type P key in picking mode to produce a postscript file.");

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete examinerViewer;
  delete action;

#ifdef SoFreePackage
  HEPVis_SoImage::clearClass();
  SoGL2PSAction::clearClass();

  SoGui::clear();  
#endif

  return true;
}
void keystroke(void*,SoEventCallback* eventCB){
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(event,P)) {
    if(examinerViewer) {
      SoDebugError::postInfo("Image","produce out.ps...");

      SoGL2PSAction* action = 
        (SoGL2PSAction*)examinerViewer->getGLRenderAction();
      action->enableFileWriting();
      examinerViewer->render();
      action->disableFileWriting();

      eventCB->setHandled();
    }
  }
}

SbString  ex1(){
  return  "\\[\\left(\\!\\!\\!\\  \\begin{array}{c}\\    n \\\\    r  \\end{array}  \\!\\!\\!\\right) = {^n}C_r = \\frac{n!}{r!(n-r)!}\\]\n";
}

SbString  ex2(){
  SbString tex;
  tex += "On peut insrer des matrices: ";
  tex += "$\\pmatrix{1&0&0\\cr 0&a&0\\cr 0&0&b\\cr}$; ";
  tex += "on peut placer des fractions, telles que ${1\\over z}$, ";
  tex += "${1\\over\\displaystyle 1+{1\\over x}}$; ";
  tex += "utiliser des lettres grecques $\\alpha$, $\\beta$, $\\gamma$, $\\Gamma$, $\\varphi$; ";
  tex += "centrer des formules complexes:";
  tex += "$$\\left|{1\\over N}\\sum_{n=1}^N \\gamma(u_n)-{1\\over 2\\pi}\\int_0^{2\\pi}\\gamma(t){\\rm d}t\\right| ";
  tex += "\\le {\\varepsilon\\over 3}.$$";
  return tex;
}

SbString  ex3(){
  return "\\Large \\begin{displaymath} \\textcolor{yellow}{=} i\\gamma^\\mu \\partial _ \\mu \\psi \\textcolor{green}{=} m \\psi \\end{displaymath}";
}
