#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/SbMath.h>
#include <HEPVis/nodes/SoPcon.h>

#include <stdlib.h>

//
// Test of SoPcon class, after JB
//
// This demo draws a Pcon
//
 
int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoPcon::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  // Set up the material for the object to be drawn

  SoMaterial* RedPlastic = new SoMaterial;
  RedPlastic->diffuseColor.setValue(1, 0, 0);
  root->addChild(RedPlastic);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  SoTimeCounter *TheCounter = new SoTimeCounter;
  TheCounter->max  = short(2*FM_PI);
  TheCounter->min = short(FM_PI/4); 
  TheCounter->frequency=0.2F;
  
  // Make the Pcon
  
  const float rmin[3]={0.2F,0.5F,0.8F}, 
    rmax[3]={1.2F,1,1}, z[3]={0,1,-1};

  SoPcon *ThePcon   = new SoPcon();
  ThePcon->setPconPts(3,0,2*FM_PI,&z[0],&rmin[0],&rmax[0]);
  ThePcon->deltaAngle = 1.5F*FM_PI;
  ThePcon->center = SbVec3f(0,0,0);

  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP"))  ThePcon->generateAlternateRep(0);
  ThePcon->deltaAngle.connectFrom(&TheCounter->output);
  root->addChild(ThePcon);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer
  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->viewAll();
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

#ifdef SoFreePackage //Cleanup
  root->unref();
  delete WriteAction;
  delete myViewer;
  SoPcon::clearClass();
  SoGui::clear();  
  printf("Pcon : exiting.\n");
#endif

  return 0; 
}

