#include <HEPVis/Gui/SoGui.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/errors/SoDebugError.h>

// HEPVis :
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodekits/SoCanvas.h>
#include <HEPVis/nodekits/SoRulerRegion.h>
#include <HEPVis/nodekits/SoDisplayRegion.h>
#include <HEPVis/actions/SoGL2PSAction.h>

extern "C" {
  void HEPVisUtils_initClasses();
}

#include <stdlib.h>

void keystroke(void*,SoEventCallback*);

SoGuiExaminerViewer* examinerViewer = NULL;

//////////////////////////////////////////////////////////////////////////////
int main (
 int //aArgc
,char** aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoGuiWidget myWindow = SoGui::init(aArgs[0]);
  if (myWindow==NULL) exit(1);

  HEPVisUtils_initClasses();

  SoCanvas* canvas = new SoCanvas;
  canvas->ref();

  SoSeparator* sceneGraph = canvas->getTopSeparator();

  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback
    (SoKeyboardEvent::getClassTypeId(),keystroke,0);
  sceneGraph->addChild(eventCallback);                                  

  // Size of the display region :
  float w = 0.9F;
  float h = 0.9F;

  // Display region :
  SoDisplayRegion* displayRegion = new SoDisplayRegion;
 {SoViewportRegion* viewportRegion = displayRegion->getViewportRegion();
  viewportRegion->setPositionPercent(1-w,1-h);
  viewportRegion->setSizePercent(w,h);
  sceneGraph->addChild(displayRegion);}

  // Horizontal ruler :
  SoRulerRegion* hRuler = new SoRulerRegion;
 {SoViewportRegion* viewportRegion = hRuler->getViewportRegion();
  viewportRegion->setPositionPercent(1-w,0);
  viewportRegion->setSizePercent(w,1-h);
  sceneGraph->addChild(hRuler);}

  // Vertical ruler :
  SoRulerRegion* vRuler = new SoRulerRegion;
 {SoViewportRegion* viewportRegion = vRuler->getViewportRegion();
  viewportRegion->setPositionPercent(0,1-h);
  viewportRegion->setSizePercent(1-w,h);
  sceneGraph->addChild(vRuler);}

  // Visualize something in the display region :
  SoSeparator* staticScene = displayRegion->getStaticScene();
  staticScene->addChild(new SoCube);

  // Connect the display region to the rulers :
  hRuler->setMeasuredRegion(displayRegion);
  vRuler->setMeasuredRegion(displayRegion);

  // View the new object in an Examiner Viewer

  examinerViewer = new SoGuiExaminerViewer(myWindow);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSize(SbVec2s(400,400));
  examinerViewer->setSceneGraph(canvas);
  examinerViewer->setCamera(displayRegion->getCamera());
  //examinerViewer->viewAll();
  examinerViewer->show();
  examinerViewer->setBackgroundColor(SbColor(0.6F,0.6F,0.6F));

  SoDebugError::postInfo("Image",
    "type P key in picking mode to produce a postscript file.");

  SoGui::show(myWindow);
  SoGui::mainLoop();

  canvas->unref();
  delete examinerViewer;
  delete action;

  return EXIT_SUCCESS;
}
//////////////////////////////////////////////////////////////////////////////
void keystroke( 
 void*
,SoEventCallback* eventCB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(event,P)) {
    if(examinerViewer) {
      SoDebugError::postInfo("Image","produce out.ps...");

      SoGL2PSAction* action = 
        (SoGL2PSAction*)examinerViewer->getGLRenderAction();
      action->enableFileWriting();
      examinerViewer->render();
      action->disableFileWriting();

      eventCB->setHandled();
    }
  }
}

