#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoOrthographicCamera.h>

#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodes/SoBackPrimitive.h>
#include <HEPVis/nodes/SoGrid.h>

#include <HEPVis/nodekits/SoShell.h>
#include <HEPVis/nodekits/SoForm.h>
#include <HEPVis/nodekits/SoPushButton.h>
#include <HEPVis/nodekits/SoList.h>

//////////////////////////////////////////////////////////////////////////////
int main(
 int
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoGuiWidget myWindow=SoGui::init(aArgv[0]);
  if(myWindow==NULL) exit(1);

  SoViewportRegion::initClass();
  SoGrid::initClass();
  SoBackPrimitive::initClass();

  SoShell::initClass();
  SoPrimitive::initClass();
  SoForm::initClass();
  SoPushButton::initClass();
  SoList::initClass();

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoViewportRegion* viewportRegion = new SoViewportRegion;
  //viewportRegion->size.setValue(SbVec2f(1,1));
  sceneGraph->addChild(viewportRegion);

  SoOrthographicCamera* camera = new SoOrthographicCamera;
  camera->position.setValue(0,0,4);
  camera->nearDistance.setValue(1);
  camera->farDistance.setValue(100);
  sceneGraph->addChild(camera);

  SoLightModel* lightModel = new SoLightModel;
  //lightModel->model.setValue(SoLightModel::BASE_COLOR);
  sceneGraph->addChild(lightModel);

  SoDirectionalLight* light = new SoDirectionalLight;
  light->color.setValue(1,1,1);
  //light->direction.setValue(SbVec3f(0.5F,-0.5F,-1));
  light->direction.setValue(SbVec3f(1,-1,-1));
  sceneGraph->addChild(light);

  SoShell* shell = new SoShell;
  sceneGraph->addChild(shell);

  SoForm* form = new SoForm;
  form->background.setValue(1,0,0);
  shell->addChild(form);

  SoPushButton* pushButton; 
  SoPushButton* prevNode; 
  // Exit :
#define BUTTON_Exit 1
  pushButton = new SoPushButton;
  pushButton->background.setValue(0.6F,0.6F,0.6F);
  pushButton->shadowThickness.setValue(0.1F);
  pushButton->marginWidth.setValue(0.1F);
  pushButton->marginHeight.setValue(0.1F);
  pushButton->labelString.setValue("Exit");
  form->addChild(pushButton);
  prevNode = pushButton;

  // Next :
#define BUTTON_Next 2
  pushButton = new SoPushButton;
  pushButton->background.setValue(0.6F,0.6F,0.6F);
  pushButton->shadowThickness.setValue(0.1F);
  pushButton->marginWidth.setValue(0.1F);
  pushButton->marginHeight.setValue(0.1F);
  pushButton->labelString.setValue("Quit");
  pushButton->leftAttachment.setValue(SoPrimitive::ATTACH_NODE);
  pushButton->leftNode.setValue(prevNode);
  form->addChild(pushButton);

  SoGuiExaminerViewer* viewer = new SoGuiExaminerViewer(myWindow);
  viewer->setSceneGraph(sceneGraph);
  viewer->setSize(SbVec2s(650,200));
  viewer->setDecoration(FALSE);
  viewer->setHeadlight(FALSE);
  viewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  sceneGraph->unref();
  delete viewer;

#ifdef SoFreePackage //Cleanup
  SoShell::clearClass();
  SoPrimitive::clearClass();
  SoForm::clearClass();
  SoPushButton::clearClass();
  SoList::clearClass();

  SoViewportRegion::clearClass();
  SoGrid::clearClass();
  SoBackPrimitive::clearClass();

  SoGui::clear();  
#endif

  return EXIT_SUCCESS; 

}
