#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/errors/SoDebugError.h>

#include <HEPVis/nodes/SoQuad.h>
#include <HEPVis/nodes/SoTriangle.h>
#include <HEPVis/nodes/So2DArrow.h>
#include <HEPVis/nodes/SoDisk.h>
#include <HEPVis/nodes/SoRing.h>
#include <HEPVis/nodes/SoCircleArc.h>
#include <HEPVis/nodes/SoPolygon.h>
#include <HEPVis/actions/SoGL2PSAction.h>

#include <stdlib.h>

void keystroke(void*,SoEventCallback*);

SoGuiExaminerViewer* examinerViewer = NULL;

int main(int,char** argv) {

  // Set up the main window
  
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  SoQuad::initClass();
  SoTriangle::initClass();
  So2DArrow::initClass();
  SoDisk::initClass();
  SoRing::initClass();
  SoCircleArc::initClass();
  SoPolygon::initClass();
  SoGL2PSAction::initClass(); //After nodes.

  SoSeparator * root = new SoSeparator;
  root->ref();

  SoShapeHints * hints = new SoShapeHints;
  hints->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
  hints->shapeType = SoShapeHints::SOLID;
  hints->creaseAngle = 0.91f;
  root->addChild( hints );


  SoSeparator * quadSep = new SoSeparator;
  root->addChild(quadSep);
  SoTransform* quadXForm = new SoTransform;
  quadXForm->translation.setValue(-5.,0.,0.);
  quadSep->addChild(quadXForm);
  SoQuad* quad = new SoQuad;
  quad->LLCornerOffset.setValue(-1.,-1.);
  quadSep->addChild(quad);


  SoSeparator * triSep = new SoSeparator;
  root->addChild(triSep);
  SoTransform* triXForm = new SoTransform;
  triXForm->translation.setValue(5.,0.,0.);
  triSep->addChild(triXForm);
  SoTriangle* triangle = new SoTriangle;
  triangle->Corner2Offset.setValue(1.,0.);
  triSep->addChild(triangle);

  
  So2DArrow* arrow = new So2DArrow;
  root->addChild(arrow);
  


  SoDisk* disk = new SoDisk;
  disk->sweepAngle.setValue(230.f);
  disk->center.setValue(0.2f,1.5f);
  root->addChild(disk);


  SoRing* ring = new SoRing;
  ring->sweepAngle.setValue(210.f);
  ring->center.setValue(0.2f,-1.5f);
  root->addChild(ring);


  SoCircleArc* arc = new SoCircleArc;
  arc->sweepAngle.setValue(120.);
  arc->center.setValue(-2.,0,0);
  root->addChild(arc);


  
  SoSeparator * polySep = new SoSeparator;
  root->addChild(polySep);
  SoTransform* polyXForm = new SoTransform;
  polyXForm->translation.setValue(-5.,5.,0.);
  polySep->addChild(polyXForm);
  SoPolygon* poly = new SoPolygon;
  polySep->addChild(poly);
  
  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                  keystroke,0);
  root->addChild(eventCallback);                                  

  // View the new object in an Examiner Viewer

  examinerViewer = new SoGuiExaminerViewer(myWindow);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSize(SbVec2s(400,400));
  examinerViewer->setSceneGraph(root);
  examinerViewer->viewAll();
  examinerViewer->setViewing(FALSE);
  examinerViewer->show();

  SoDebugError::postInfo("Test2D",
    "type P key in picking mode to produce a postscript file.");

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete examinerViewer;
  delete action;

  return 0;
}
void keystroke(void*,SoEventCallback* eventCB){
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(event,P)) {
    if(examinerViewer) {
      SoDebugError::postInfo("Test2D","produce out.ps...");

      SoGL2PSAction* action = 
        (SoGL2PSAction*)examinerViewer->getGLRenderAction();
      action->enableFileWriting();
      examinerViewer->render();
      action->disableFileWriting();

      eventCB->setHandled();
    }
  }
}
