#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/actions/SoWriteAction.h>
#include <HEPVis/nodes/SoTorus.h>

//
// Test of SoTorus class, after Torus by JB
//
// This demo draws a Torus
//
 
int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoTorus::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  // Set up the material for the object to be drawn

  SoMaterial    *RedPlastic = new SoMaterial;
  RedPlastic->diffuseColor.setValue(1, 0, 0);
  root->addChild(RedPlastic);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  SoTimeCounter *TheCounter = new SoTimeCounter;
  TheCounter->max  = 150; 
  TheCounter->min = 50;
  TheCounter->frequency=0.2F;

  // Make the Torus

  SoTorus *TheTorus   = new SoTorus();
  TheTorus->rMajor   = 100;
  TheTorus->rMinor   = 20;
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP"))  TheTorus->generateAlternateRep(0);
  TheTorus->rMajor.connectFrom(&TheCounter->output);
  root->addChild(TheTorus);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer
  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->viewAll();
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoTorus::clearClass();
  SoGui::clear();  
#endif

  return 0; 
}

