#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/nodes/SoVtxReconTrack.h>
#include <HEPVis/nodes/SoEllipticalPrism.h>
#include <HEPVis/nodes/SoEllipsoid.h>

// Test of SoVtxReconTrack class
// Joe Boudreau Nov. 7, 1996

// This demo draws a handfull of tracks:

#define NTRACKS 6
float TrackParameters[NTRACKS][5]= {
  {0.0235598F,-2.64028F,0.223126F,-0.000947721F,-19.2387F},
  {-0.026806F,-1.85632F,3.03883F,0.000792523F,-19.2367F},
  {0.0226147F,0.517672F,1.25412F,-0.000769267F,-19.2382F},
  {-0.0102705F,-1.00757F,1.10671F,-0.0008555F,-19.2373F},
  {-0.00616535F,-1.37573F,1.55067F,-0.000547534F,-19.2367F},
  {0.0106418F,-2.46822F,4.55818F,0.000654988F,-19.2397F},
};


 
int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoVtxReconTrack::initClass();
  SoEllipsoid::initClass();
  SoEllipticalPrism::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();
  // Set up the light model and the material for
  // The object to be drawn

  SoMaterial    *TrackMaterial = new SoMaterial;
  TrackMaterial->diffuseColor.setValue(0,1,1);
  root->addChild(TrackMaterial);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  SoElapsedTime *TheCounter = new SoElapsedTime;
  SoCalculator *TheCalculator = new SoCalculator();
  TheCalculator->a.connectFrom(&TheCounter->timeOut);
  // TheCalculator->expression.set1Value(0,"oa=0.1*sin(a)*sin(a)+0.1");
  TheCalculator->expression.setValue("oA=vec3f(0.10*sin(a)*sin(a)+0.1,0.10*cos(a)*cos(a)+0.1,0.1)");


  // Make the Tracks
  SoEllipsoid* AnEllipsoid = new SoEllipsoid();
  AnEllipsoid->center.setValue(0,0,-19.5F);
  AnEllipsoid->eigenvalues.connectFrom(&TheCalculator->oA);
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) AnEllipsoid->generateAlternateRep(0);

  for (int i=0;i<NTRACKS;i++) {
    SoVtxReconTrack* TheVtxReconTrack = new SoVtxReconTrack;
    TheVtxReconTrack->cotTheta.setValue(TrackParameters[i][1]);
    TheVtxReconTrack->phi0.setValue(TrackParameters[i][2]);
    TheVtxReconTrack->ellipsoid.setValue(AnEllipsoid);
    TheVtxReconTrack->s0.setValue(0.5F);
    TheVtxReconTrack->s1.setValue(7);
  
    // Generate the alternate representation to generate an inventor
    // file readable by anyone, and write it out. This is an important
    // feature of any new HepVis node.

    if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) TheVtxReconTrack->generateAlternateRep(0);
    root->addChild(TheVtxReconTrack);
  }
  root->addChild(AnEllipsoid);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer
  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoVtxReconTrack::clearClass();
  SoEllipsoid::clearClass();
  SoEllipticalPrism::clearClass();
  SoGui::clear();  
#endif

  return 0; 
}

