/////////////////////////////////////////////////////////////
//
// Simple example of creation of an Twig tree.
// The Twigs, which display a cone and a sphere, are 
// derived in TestTwigs.C
//
//
/////////////////////////////////////////////////////////////

#include <math.h>
#include <stdlib.h>

#include <iostream>

#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSeparator.h>

#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>

#include <HEPVis/Xt/Twig.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCube.h>
#include "TestTwigs.h"

int  main(int, char ** argv);

SoSeparator   *root;   
Widget         myWindow;

SoXtExaminerViewer *myViewer = NULL;

int main(int, char ** argv)
{
          
  myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) {
    std::cout << std::endl << "SoXt::init failed to initialise X11 => exit" << std::endl; 
    exit(1);}
  
  SoSelection *root = new SoSelection();
  root->ref();
  
  
///////////////////////////////////////////////////////////////
//
//   Create the twigs
//   ================
//
///////////////////////////////////////////////////////////////
 

 Twig         *frameTwig         = new Twig ("Frame 0", TRUE);
 Twig         *eventTwig         = new Twig ("Event" ,  TRUE, TRUE);
 
 // ConeTwig, SphereTwig and BoxTwig are derived from Twig.  See code in TestTwigs.
 //
 ConeTwig     *theConeTwig       = new ConeTwig   ("Cone",   TRUE );
 SphereTwig   *theSphereTwig     = new SphereTwig ("Sphere", TRUE );
 BoxTwig      *theBoxTwig        = new BoxTwig    ("Boxes", TRUE  );

///////////////////////////////////////////////////////////////
//
//   Attach the top SoSeparator in twig tree to root SoSeparator
//   ===========================================================
// 
///////////////////////////////////////////////////////////////

 root->addChild (frameTwig->getIvNode());

///////////////////////////////////////////////////////////////
//
//   Arrange twigs to form a tree
//   ============================
// 
///////////////////////////////////////////////////////////////

 frameTwig -> addChild (eventTwig); 
 
   eventTwig -> addChild (theConeTwig);  
   eventTwig -> addChild (theSphereTwig);
   eventTwig -> addChild (theBoxTwig);    

///////////////////////////////////////////////////////////////
//
//   Make the Motif widgets associated with the twig tree
//   ====================================================
// 
///////////////////////////////////////////////////////////////

 eventTwig->setParentWidget(myWindow);

 // normally, one would invoke createShowDialog (see below)
 // in a callback function attached to some button so that
 // if the widget is destroyed, the user can created it again.
 //
 eventTwig->createShowDialog();
 frameTwig->drawTree();
 
///////////////////////////////////////////////////////////////
//
//   Setup for Picking
//   =================
//
///////////////////////////////////////////////////////////////
   TwigCallbackStruct *theCBStruct = new TwigCallbackStruct(frameTwig);
   // Enable select callbacks
   root->addSelectionCallback(Twig::selectionCB, static_cast<void *> (theCBStruct)); 
 
///////////////////////////////////////////////////////////////
//
//   Create the viewer
//   =================
// 
///////////////////////////////////////////////////////////////

 myViewer = new SoXtExaminerViewer(myWindow); 
 myViewer->setSceneGraph(root);
 myViewer->show();
 
 SoXt::show(myWindow);   
 
 SoXt::mainLoop();     // we never return from here...
    
 return 0;
}

void ConeTwig::drawMe() {
  Twig::drawMe();

  // make the shape only if it hasn't already been made
  if (okToRedraw()) {  

    // Define material
    // ---------------

    SoMaterial *copper = new SoMaterial;
    copper->ambientColor.setValue (0.40, 0.20, 0.30);
    copper->diffuseColor.setValue (0.80, 0.55, 0.20);
    copper->specularColor.setValue(0.40, 0.30, 0.10);
    copper->emissiveColor.setValue(0.25, 0.15, 0.15); 
    copper->shininess = 1.0;
    copper->transparency = 0.0;

    SoCone *myCone = new SoCone;
    
    oiv_node->addChild(copper);
    oiv_node->addChild(myCone);
    
  }
}


void SphereTwig::drawMe() {
  Twig::drawMe(); 

  if (okToRedraw()) {
 
    // Define material
    // ---------------

    SoMaterial *steel = new SoMaterial;
    steel->ambientColor.setValue (0.30, 0.20, 0.40);
    steel->diffuseColor.setValue (0.20, 0.55, 0.80);
    steel->specularColor.setValue(0.10, 0.30, 0.40);
    steel->emissiveColor.setValue(0.15, 0.15, 0.25);
    steel->shininess = 1.0;
    steel->transparency = 0.0;
    
    SoSphere *mySphere = new SoSphere;
    
    oiv_node->addChild(steel);
    oiv_node->addChild(mySphere);
       
  }
}

void BoxTwig::drawMe() {
  Twig::drawMe(); 

  if (okToRedraw()) {
 
    // Define material
    // ---------------

    SoMaterial *steel = new SoMaterial;
    steel->ambientColor.setValue (0.30, 0.20, 0.40);
    steel->diffuseColor.setValue (0.20, 0.55, 0.80);
    steel->specularColor.setValue(0.10, 0.30, 0.40);
    steel->emissiveColor.setValue(0.15, 0.15, 0.25);
    steel->shininess = 1.0;
    steel->transparency = 0.0;
    
    SoCube *myBox = new SoCube;
    
    oiv_node->addChild(steel);
    oiv_node->addChild(myBox);
       
  }
}














