
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoSeparator.h>

#include "../source/iv2pov.h"

#include <HEPVis/SbArguments.h>

//G.Barrand :
#ifdef HEPVIS_READ_HEPVIS
#include <Inventor/nodekits/SoNodeKit.h>
extern "C" {
  void HEPVisUtils_initClasses();
  void HEPVisGeometry_initClasses();
  void HEPVisDetector_initClasses();
  void HEPVisPlotter_initClasses();
  void HEPVisGUI_initClasses();
}
#endif

#include <iostream>

//////////////////////////////////////////////////////////////////////////////
int main( 
 int aArgc
,char* aArgv[]
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SbArguments args(aArgc,aArgv);
  if((aArgc==1)||args.isAnArgument("-h")) {
    std::cout << "hiv2pov :" << std::endl;
    std::cout << "Usage :" << std::endl;
    std::cout << " OS> hiv2pov <inventor file>" << std::endl;
    std::cout << "If successfull, an out.pov is produced." << std::endl;
    return 0;
  }

  ///////////////////////////////////////////////
  // Args :
  ///////////////////////////////////////////////

  SbBool rmBack = FALSE;
  //SbString arg_main_hiv;
  //if(args.find("-b",arg_main_hiv)==FALSE) {
  //}

  SbString file;
  if(args.file(file)==FALSE) {
    return 1;
  }

  SoDB::init();

#ifdef HEPVIS_READ_HEPVIS
  SoNodeKit::init();
  HEPVisUtils_initClasses();
  HEPVisGeometry_initClasses();
  HEPVisDetector_initClasses();
  HEPVisPlotter_initClasses();
  HEPVisGUI_initClasses();
#endif

  SoInput in;
  if(!in.openFile(file.getString()))  {
    fprintf(stderr, "hiv2pov : Cannot open %s\n",file.getString());
    return 1;
  }
  SoSeparator* root = SoDB::readAll(&in);
  if(root == NULL) {
    fprintf(stderr, "hiv2pov : Problem reading data in %s\n",
      file.getString());
    return 1;
  }
  in.closeFile();

  root->ref();

  //NOTE : no need to produce alternate reps since 
  //       iv2pov works on shape triangles.

  if(HEPVis_iv2pov(root,TRUE,rmBack,"out.pov")==FALSE) {
    root->unref();
    return 1;
  }

  root->unref();

  return 0;
}
