/* 
 *  SbTexToBitmap.cxx
 *  
 *
 *  Created by Laurent Garnier on Wen May 18 2005.
 *
 */ 

// this :
#include <HEPVis/SbLatex.h>

#include <HEPVis/SbFile.h>
#include <HEPVis/SbString.h>

#include <stdio.h>
#include <stdlib.h>

SbBool SbLatex::fileToDvi (const SbString& aFile,SbString& aDviFile,const SbString& aDviGenCommand)
{

  // remove extention if exist
  SbString filename = aFile;
  int pos = SbStringReverseFind(filename,".tex");
  if (pos != SbStringNotFound) filename.deleteSubString(pos);

  aDviFile = "";
  SbString tmpDir;
  if (!changeToTmpDir(tmpDir)) // "cd" command has no effect on latex command
                               // but this is use to get the tmp directory
    return FALSE;

  SbString shortName = tmpDir;
#ifdef WIN32
  pos = SbStringReverseFind(filename,"\\");
  if (pos != SbStringNotFound) {
    shortName += filename.getSubString(pos+1);
  } else {
    shortName = filename;
  }
#else
  pos = SbStringReverseFind(filename,"/");
  if (pos != SbStringNotFound) {
    shortName += filename.getSubString(pos+1);
  } else {
    shortName = filename;
  }
#endif
  SbString auxName = shortName;
  auxName += ".aux";
  SbString logName = shortName;
  logName += ".log";
  SbString nullName = shortName;
  nullName += "Null";

  SbString dviCommand = "cd ";
  dviCommand +=tmpDir;
  dviCommand +="; ";
  pos = SbStringFind(aDviGenCommand,"%M");
  if (pos!=-1) {
    dviCommand += aDviGenCommand.getSubString(0,pos-1);
    if (SbStringFind(aFile,"..") == 0) {
      dviCommand +=fCurrentDir;
#ifdef WIN32
    } else if (SbStringFind(aFile,":\\") != 1) {
      dviCommand +=fCurrentDir;
#else
    } else if (SbStringFind(aFile,"/") != 0) {
      dviCommand +=fCurrentDir;
#endif
    }
    dviCommand +=aFile;
    pos = SbStringFind(aDviGenCommand,"%M");
    dviCommand +=aDviGenCommand.getSubString(pos+2);
    dviCommand +=nullName;
  } else {
    return FALSE;
  }
  // execute latex command
  if(::system(dviCommand.getString())){}
  ::remove(auxName.getString());
  ::remove(logName.getString());
  ::remove(nullName.getString());

  if (!changeToCurrentDir())
    return FALSE;
  aDviFile =shortName;
  aDviFile += ".dvi";
  return TRUE;
}
  

SbBool SbLatex::strToDvi (const SbString& aTexStr,SbString& aDviFile,const SbString& aTmpFile)
{
  SbString tmpDir;
  if (!changeToTmpDir(tmpDir))
    return FALSE;
  SbString texFile = tmpDir;
  texFile += aTmpFile;
  FILE *fp = fopen (texFile.getString(), "wb");
  fwrite(aTexStr.getString(), 1, aTexStr.getLength(), fp );
  fclose(fp);
  if (!changeToCurrentDir())
    return FALSE;
  if (!fileToDvi(texFile,aDviFile))
    return FALSE;
  if (!changeToTmpDir(tmpDir))
    return FALSE;
  remove(texFile.getString());
  if (!changeToCurrentDir())
    return FALSE;

  return TRUE;
}

SbBool SbLatex::changeToTmpDir(SbString& aTmpDir)
{
  aTmpDir ="";
  const char* cDir = SbFile::getWorkingDirectory();
  fCurrentDir = SbString(cDir);
  if (cDir == NULL)
    return FALSE;
  if(!SbFile::tmpname(aTmpDir,"","",""))
    return FALSE;
#ifdef WIN32
  aTmpDir = aTmpDir.getSubString(0,SbStringReverseFind(aTmpDir,"\\"));
  if (SbStringReverseFind(fCurrentDir,"\\") != (fCurrentDir.getLength()-1)) {
    fCurrentDir += "\\";
  }
#else
  aTmpDir = aTmpDir.getSubString(0,SbStringReverseFind(aTmpDir,"/"));
  if (SbStringReverseFind(fCurrentDir,"/") != (fCurrentDir.getLength()-1)) {
    fCurrentDir += "/";
  }
#endif
  SbFile sbFile;
  if (sbFile.openDirectory(aTmpDir.getString()) == NULL) {
    return FALSE;
  }
  if (aTmpDir == "")
    return FALSE;
  SbString cdCommand = "cd ";
  cdCommand += aTmpDir;
  if (system(cdCommand.getString()) == -1)
    return FALSE;
  return TRUE;
}
SbBool SbLatex::changeToCurrentDir() const
{
  SbFile sbFile;
  if (sbFile.openDirectory(fCurrentDir.getString()) == NULL) {
    return FALSE;
  }
  SbString cdCommand = "cd ";
  cdCommand += fCurrentDir;
  if (system(cdCommand.getString()) == -1)
    return FALSE;
  return TRUE;
}
