//#define DEBUG

// this :
#include <HEPVis/Xt/SbPainterX11.h>

#include <Inventor/errors/SoDebugError.h>

#ifdef HEPVisXt

static float ConvertRGB_ToGrey(float,float,float);

#define DBL_INT (int)(float)
#define FLOAT_SHORT (short)(float)
#define MAX_SHORT 32768.
#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
SbPainterX11::SbPainterX11(
)
:fDisplay(NULL)
,fColormap(0L)
,fVclass(0)
,fDrawable(0L)
,fPointn(0)
,fPoints(NULL)
,fSegmentn(0)
,fSegments(NULL)
,fWarnDoneColor(FALSE)
,fWarnDoneShort(FALSE)
,fColori(0)
,fPixel(0L)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
SbPainterX11::~SbPainterX11(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete [] fPoints;
  delete [] fSegments;

  clearColors();

  fDisplay = NULL;
  fColormap = 0L;
  fVclass = 0;
  fDrawable = 0L;
  fWarnDoneColor = FALSE;
  fWarnDoneShort = FALSE;
  fPoints = NULL;
  fPointn = 0;
  fSegments = NULL;
  fSegmentn = 0;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::beginTraversal(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::endTraversal(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::setWindow(
 Display* aDisplay
,Drawable aDrawable
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aDisplay==NULL) return;
  if(aDrawable==0L) return;

  fAtb.initialize(aDisplay);

  clearColors();

  fDisplay = aDisplay;
  fDrawable = aDrawable;

  XWindowAttributes watbs;
  XGetWindowAttributes(fDisplay,fDrawable,&watbs);
  fColormap = watbs.colormap;
#ifdef __cplusplus
  fVclass = watbs.visual->c_class;
#else
  fVclass = watbs.visual->class;
#endif

}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::clearColorBuffer(
 float aRed
,float aGreen
,float aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fRed = aRed;
  fGreen = aGreen;
  fBlue = aBlue;
  fPixel = convertRGB_ToPixel(fRed,fGreen,fBlue);
  //
  if(fDisplay==NULL) return;
  if(fDrawable==0L)  return;
  GC gc = fAtb.getGC();
  if(gc==NULL) return;
  fAtb.reset();
  fAtb.setColor(fPixel);
  fAtb.setAreaStyle(SbAreaSolid);
  fAtb.flush();
  XFillRectangle(fDisplay,fDrawable,gc,0,0,fWindowWidth,fWindowHeight);
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::drawPrimitive (
 SbPrimitiveType aType
,int aPointn
,float* aXs
,float* aYs
,float* aZs
,const SbPainterContext& aAtb
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(allocatePoints(aType,aPointn,aAtb)==FALSE) return;
  unsigned long pixel = 
    convertRGB_ToPixel(aAtb.fRed,aAtb.fGreen,aAtb.fBlue);
  switch(aType) {
  case SbPrimitivePoints:
    if(convertLineToPoints(aPointn,aXs,aYs,aZs)==TRUE) {
      drawMarkers(aPointn,fPoints,pixel,aAtb.fMarkerStyle,aAtb.fMarkerSize);
    }
    break;
  case SbPrimitiveLineStrip:
  case SbPrimitiveLineLoop:
    if(convertLineToPoints(aPointn,aXs,aYs,aZs)==TRUE) {
      drawLines(aPointn,fPoints,pixel,aAtb.fLinePattern,aAtb.fLineWidth);
    }
    break;
  case SbPrimitivePolygon:
    if(convertLineToPoints(aPointn,aXs,aYs,aZs)==TRUE) {
      drawPolygon(aPointn,fPoints,pixel,aAtb.fAreaStyle);
    }
    break;
  case SbPrimitiveLines:
    break;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
unsigned long SbPainterX11::convertRGB_ToPixel(
 float aRed
,float aGreen
,float aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDisplay==NULL) return 0L;

  SbBool found = FALSE;
  int count;
  for(count=0;count<fColori;count++) {
    if ( (aRed == fColors[count].fRed) &&
	 (aGreen == fColors[count].fGreen) &&
	 (aBlue == fColors[count].fBlue) ) {
      found = TRUE;
      break;
    }
  }
  if(found==TRUE) return fColors[count].fPixel;

  unsigned long pixel;
  SbBool allocated;
  SbBool status = getPixelFromRGB(aRed,aGreen,aBlue,pixel,allocated);
  if(status==FALSE) {
    if(fWarnDoneColor==FALSE) {
      SoDebugError::postInfo("SbPainterX11::convertRGB_ToPixel",
        "unable to find a pixel.\n");
      fWarnDoneColor = TRUE;
    }
  }

  if(fColori>=MAX_COLORS) {
    SoDebugError::postInfo("SbPainterX11::convertRGB_ToPixel",
      "buffer too short.\n");
    return pixel;
  }
  fColors[fColori].fRed = aRed;  
  fColors[fColori].fGreen = aGreen;
  fColors[fColori].fBlue = aBlue;
  fColors[fColori].fPixel = pixel;
  fColors[fColori].fAllocated = allocated;
  fColori ++;
  return pixel;
}
//////////////////////////////////////////////////////////////////////////////
SbBool SbPainterX11::convertPixelToRGB(
 unsigned long aPixel
,float& aRed
,float& aGreen
,float& aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  for(int count=0;count<fColori;count++) {
    if(aPixel==fColors[count].fPixel) {
      aRed = fColors[count].fRed;
      aGreen = fColors[count].fGreen;
      aBlue  = fColors[count].fBlue;
      return TRUE;
    }
  }
  aRed = 0.;
  aGreen = 0.;
  aBlue = 0.;
  return FALSE;
}
//////////////////////////////////////////////////////////////////////////////
SbBool SbPainterX11::convertLineToPoints (
 int aNumber
,float* aXs
,float* aYs
,float* /*aZs*/
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aNumber<=0) return FALSE;
  float x = aXs[0];
  float y = aYs[0];
  float xmin = x;
  float xmax = x;
  float ymin = y;
  float ymax = y;
  if((fabs(x)>=MAX_SHORT)||(fabs(y)>=MAX_SHORT)) {
    if(fWarnDoneShort==FALSE) {
      SoDebugError::postInfo("SbPainterX11::convertLineToPoints",
        "can't convert float to short.\n");
      fWarnDoneShort = TRUE;
    }
    return FALSE;
  }
  fPoints[0].x = FLOAT_SHORT x;
  fPoints[0].y = fWindowHeight - FLOAT_SHORT y;
  for(int count=1;count<aNumber;count++) { 
    x = aXs[count];
    y = aYs[count];
    xmin = MINIMUM(xmin,x);
    xmax = MAXIMUM(xmax,x);
    ymin = MINIMUM(ymin,y);
    ymax = MAXIMUM(ymax,y);
    if((fabs(x)>=MAX_SHORT)||(fabs(y)>=MAX_SHORT)) {
      if(fWarnDoneShort==FALSE) {
	SoDebugError::postInfo("SbPainterX11::convertLineToPoints",
          "can't convert float to short.\n");
	fWarnDoneShort = TRUE;
      }
      return FALSE;
    }
    fPoints[count].x = FLOAT_SHORT x;
    fPoints[count].y = fWindowHeight - FLOAT_SHORT y;
  }
  // Out of viewport. A true clipping over edges would be better :
  if(xmin>=(float)fWindowWidth) return FALSE; 
  if(xmax <0.) return FALSE;
  if(ymin>=(float)fWindowHeight) return FALSE;
  if(ymax <0.) return FALSE;
  // Ok :
  return TRUE;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::drawPolygon(
 int aPointn
,XPoint* aPoints
,unsigned long aPixel
,SbAreaStyle aStyle
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDisplay==NULL) return;
  if(fDrawable==0L) return;
  GC gc = fAtb.getGC();
  if(gc==NULL) return;
  fAtb.reset();
  fAtb.setColor(aPixel);
  fAtb.setAreaStyle(aStyle);
  fAtb.flush();
  XFillPolygon(fDisplay,fDrawable,gc,
	       aPoints,aPointn,Nonconvex,CoordModeOrigin);
  if(fEdges==TRUE) {
    fAtb.reset();
    fAtb.setColor(fPixel);
    fAtb.setLinePattern(SbLinePattern_solid);
    fAtb.setLineWidth(1);
    fAtb.flush();
    XDrawLines(fDisplay,fDrawable,gc,aPoints,aPointn,CoordModeOrigin);
  }
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::drawLines(
 int aPointn
,XPoint* aPoints
,unsigned long aPixel
,SbLinePattern aPattern
,int aWidth 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDisplay==NULL) return;
  if(fDrawable==0L) return;
  GC gc = fAtb.getGC();
  if(gc==NULL) return;
  fAtb.reset();
  fAtb.setColor(aPixel);
  fAtb.setLinePattern(aPattern);
  fAtb.setLineWidth(aWidth);
  fAtb.flush();
  if( (aPointn==2) &&
      (aPoints[0].x == (int)aPoints[1].x) &&
      (aPoints[0].y == (int)aPoints[1].y) ) {
    XDrawPoint(fDisplay,fDrawable,gc,(int)aPoints[0].x,(int)aPoints[0].y);
  } else {
    XDrawLines(fDisplay,fDrawable,gc,aPoints,aPointn,CoordModeOrigin);
  }
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::drawMarkers (
 int aPointn
,XPoint* aPoints
,unsigned long aPixel
,SbMarkerStyle aStyle 
,int aSize 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDisplay==NULL) return;
  if(fDrawable==0L) return;
  GC gc = fAtb.getGC();
  if(gc==NULL) return;

  int mark_size  = aSize <=0 ? 4 : aSize;   
  if(aStyle==SbMarkerCircleLine) {
    fAtb.reset();
    fAtb.setColor(aPixel);
    fAtb.setLinePattern(SbLinePattern_solid);
    fAtb.setLineWidth(1);
    fAtb.flush();
    int center_x,center_y;
    for(int count=0;count<aPointn;count++) {
      center_x = aPoints[count].x - mark_size/2;
      center_y = aPoints[count].y - mark_size/2;
      XDrawArc(fDisplay,fDrawable,gc,
	       center_x,center_y,
	       (unsigned int)mark_size,(unsigned int)mark_size,0,23040);
    }
  } else if(aStyle==SbMarkerDot) { 
    fAtb.reset();
    fAtb.setColor(aPixel);
    fAtb.flush();
    XDrawPoints(fDisplay,fDrawable,gc,aPoints,aPointn,CoordModeOrigin);
  } else {
    int number = getMarkerNumberOfSegments(aStyle);
    int ipoint = 0;
    int center_x,center_y;
    for(int count=0;count<aPointn;count++) {
      center_x = aPoints[count].x;
      center_y = aPoints[count].y;
      getMarkerSegments(aStyle,mark_size,
			center_x,center_y,&(fSegments[ipoint]));
      ipoint += number;
    }
    if(ipoint==0) return;
    fAtb.reset();
    fAtb.setColor(aPixel);
    fAtb.setLinePattern(SbLinePattern_solid);
    fAtb.setLineWidth(1);
    fAtb.flush();
    XDrawSegments(fDisplay,fDrawable,gc,fSegments,ipoint);
  }
}
//////////////////////////////////////////////////////////////////////////////
int SbPainterX11::getMarkerNumberOfSegments(
 SbMarkerStyle aStyle 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  switch (aStyle) {
  case SbMarkerPlus:  return 2;
  case SbMarkerAsterisk:  return 4;
  case SbMarkerCross:  return 2;
  case SbMarkerStar:  return 3;
  case SbMarkerCircleLine:  return 0;
  case SbMarkerDot:  return 0;
  default:  return 1;
  }
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::getMarkerSegments(
 SbMarkerStyle aStyle 
,int aSize 
,int aCenter_x 
,int aCenter_y 
,XSegment* aSegment 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  XSegment* segment = (XSegment*)aSegment;
  aSize /= 2;
  aSize = aSize==0 ? 1 : aSize;
  switch(aStyle) {
  case SbMarkerPlus:
    segment[0].x1 = (short)(aCenter_x - aSize);
    segment[0].y1 = (short)aCenter_y ;
    segment[0].x2 = (short)(aCenter_x + aSize);
    segment[0].y2 = (short)aCenter_y ;
    
    segment[1].x1 = (short)aCenter_x ;
    segment[1].y1 = (short)(aCenter_y - aSize);
    segment[1].x2 = (short)aCenter_x ;
    segment[1].y2 = (short)(aCenter_y + aSize);
    break;
  case SbMarkerAsterisk:
    segment[0].x1 = (short)(aCenter_x - aSize);
    segment[0].y1 = (short)aCenter_y ;
    segment[0].x2 = (short)(aCenter_x + aSize);
    segment[0].y2 = (short)aCenter_y ;

    segment[1].x1 = (short)aCenter_x ;
    segment[1].y1 = (short)(aCenter_y - aSize);
    segment[1].x2 = (short)aCenter_x ;
    segment[1].y2 = (short)(aCenter_y + aSize);
    
    segment[2].x1 = (short)(aCenter_x - DBL_INT (aSize * .707));
    segment[2].y1 = (short)(aCenter_y - DBL_INT (aSize * .707));
    segment[2].x2 = (short)(aCenter_x + DBL_INT (aSize * .707));
    segment[2].y2 = (short)(aCenter_y + DBL_INT (aSize * .707));
    
    segment[3].x1 = (short)(aCenter_x + DBL_INT (aSize * .707));
    segment[3].y1 = (short)(aCenter_y - DBL_INT (aSize * .707));
    segment[3].x2 = (short)(aCenter_x - DBL_INT (aSize * .707));
    segment[3].y2 = (short)(aCenter_y + DBL_INT (aSize * .707));
    break;
  case SbMarkerCross:
    segment[0].x1 = (short)(aCenter_x - aSize);
    segment[0].y1 = (short)(aCenter_y - aSize);
    segment[0].x2 = (short)(aCenter_x + aSize);
    segment[0].y2 = (short)(aCenter_y + aSize);
    
    segment[1].x1 = (short)(aCenter_x + aSize);
    segment[1].y1 = (short)(aCenter_y - aSize);
    segment[1].x2 = (short)(aCenter_x - aSize);
    segment[1].y2 = (short)(aCenter_y + aSize);
    break;
  case SbMarkerStar:
    segment[0].x1 = (short)aCenter_x;
    segment[0].y1 = (short)(aCenter_y + aSize);
    segment[0].x2 = (short)(aCenter_x - DBL_INT (aSize * .6));
    segment[0].y2 = (short)(aCenter_y - DBL_INT (aSize * .5));
    
    segment[1].x1 = segment[0].x2 ;
    segment[1].y1 = segment[0].y2 ;
    segment[1].x2 = (short)(aCenter_x + DBL_INT (aSize * .6));
    segment[1].y2 = (short)(aCenter_y - DBL_INT (aSize * .5));

    segment[2].x1 = segment[1].x2;
    segment[2].y1 = segment[1].y2;
    segment[2].x2 = segment[0].x1;
    segment[2].y2 = segment[0].y1;
    break;
  default:
    segment[0].x1 = (short)aCenter_x;
    segment[0].y1 = (short)aCenter_y;
    segment[0].x2 = (short)aCenter_x;
    segment[0].y2 = (short)aCenter_y;
    break;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
SbBool SbPainterX11::getPixelFromRGB (
 float aRed
,float aGreen
,float aBlue
,unsigned long& aPixel
,SbBool& aAlloc
)
//////////////////////////////////////////////////////////////////////////////
/* read_write visual class: PseudoColor, GrayScale, DirectColor.           */
/* read_only  visual class: TrueColor, StaticColor, StaticGray.            */
/* Could allocate a read_only cell in a read_write, read_only visual */
/* Could allocate a read_write cell only in read_write visual        */
/* StaticGray + (depth=1) is monochrome                                    */
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aPixel = 0L;
  aAlloc = FALSE;
  if(fDisplay==NULL) return FALSE;
  if(fColormap==0L) return FALSE;

  SbBool is_grey = ((fVclass==GrayScale) || (fVclass==StaticGray) ) ? 
                   TRUE : FALSE;
  float grey = is_grey==TRUE ? 
    ConvertRGB_ToGrey(aRed,aGreen,aBlue) : 0.;

  XColor xcolor;
  xcolor.pixel = 0L;
  xcolor.red = (unsigned short) ((is_grey==TRUE ? grey : aRed) * 0xffff);
  xcolor.green = (unsigned short) ((is_grey==TRUE ? grey : aGreen) * 0xffff);
  xcolor.blue = (unsigned short) ((is_grey==TRUE ? grey : aBlue) * 0xffff);
  if(XAllocColor(fDisplay,fColormap,&xcolor)==0) {
    // Could be done on read_only/read_write visual.
    // Cell is taken from a common pool. It is read only.
    // Color not found. Try to allocate a private color cell.
    if( (fVclass==TrueColor)   || 
	(fVclass==StaticColor) || 
	(fVclass==StaticGray)
        ) { // Viusal class is read_only.
      return FALSE;
    } else {
      // read/write cell :
      if(XAllocColorCells(fDisplay,fColormap,False,0L,0,&(xcolor.pixel),1)==0) { 
	// Alloc fail :
	return FALSE;
      } else {
	xcolor.flags = DoRed | DoGreen | DoBlue;
	XStoreColor(fDisplay,fColormap,&xcolor);
	aPixel = xcolor.pixel;
	aAlloc = TRUE;
      }
    }
  } else {
    aPixel = xcolor.pixel;
    aAlloc = FALSE;
  }
  return TRUE;
}
//////////////////////////////////////////////////////////////////////////////
SbBool SbPainterX11::allocatePoints(
 SbPrimitiveType aType
,int aPointn
,const SbPainterContext& aAtb
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPointn==0) return FALSE;
  if(aPointn>fPointn) {
    if(fPoints!=NULL) delete [] fPoints;
    fPoints = new XPoint[aPointn];
    if(fPoints==NULL) {
      fPoints = NULL;
      fPointn = 0;
      return FALSE;
    }
    fPointn = aPointn;
  }
  
  if( (aType==SbPrimitiveLines) || (aType==SbPrimitivePoints) ) { 
    int segn = (aType==SbPrimitivePoints) ? 
      getMarkerNumberOfSegments(aAtb.fMarkerStyle) * aPointn : aPointn/2;
    if(segn>fSegmentn) {
      if(fSegments!=NULL) delete [] fSegments;
      fSegments = new XSegment[segn];
      if(fSegments==NULL) {
	fSegments = NULL;
	fSegmentn = 0;
	return FALSE;
      }
      fSegmentn  = segn;
    }
  }
  return TRUE;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterX11::clearColors(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDisplay==NULL) return;
  if(fColormap==0L) return;
  int pixeli = 0;
  unsigned long pixels[MAX_COLORS];
  for(int count=0;count<fColori;count++) {
    if(fColors[count].fAllocated==TRUE) {
      pixels[pixeli] = fColors[count].fPixel;
      pixeli ++;
    }
  }
  XFreeColors(fDisplay,fColormap,pixels,pixeli,0L);
  fColori = 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
SbXContext::SbXContext(
)
:fDisplay(NULL)
,fGC(NULL)
,fStippleHatchedPixmap(0L)
,fStippleCheckerPixmap(0L)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
SbXContext::~SbXContext(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDisplay==NULL) return;
  if(fGC!=NULL) XFreeGC(fDisplay,fGC);
  if(fStippleHatchedPixmap!=0L) 
    XFreePixmap(fDisplay,fStippleHatchedPixmap);
  if(fStippleCheckerPixmap!=0L) 
    XFreePixmap(fDisplay,fStippleCheckerPixmap);
  fStippleHatchedPixmap = 0L;
  fStippleCheckerPixmap = 0L;
  fDisplay = NULL;
  fGC = NULL;
}
//////////////////////////////////////////////////////////////////////////////
GC SbXContext::getGC(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fGC;
}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::initialize(
 Display* aDisplay
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if((fDisplay!=NULL) && (fGC!=NULL)) XFreeGC(fDisplay,fGC);
  fDisplay = NULL;
  fGC = NULL;

  if(aDisplay==NULL) return;

  fDisplay = aDisplay;

  fFlags = 0L;

  fArgs.background  = XBlackPixelOfScreen(XDefaultScreenOfDisplay(fDisplay));
  fFlags |= GCBackground;

  fPixel = XWhitePixelOfScreen(XDefaultScreenOfDisplay(fDisplay));
  fArgs.foreground  = XWhitePixelOfScreen(XDefaultScreenOfDisplay(fDisplay));
  fFlags |= GCForeground;

  fLineWidth = 0;
  fArgs.line_width = 0;
  fFlags |= GCLineWidth;

  fLinePattern = SbLinePattern_solid;
  fArgs.line_style = LineSolid;
  fFlags |= GCLineStyle;

  fAreaStyle = SbAreaSolid;
  fArgs.fill_style = FillSolid;
  fFlags |= GCFillStyle;

  fGC = (GC) XCreateGC(fDisplay,
		      XDefaultRootWindow(fDisplay),
		      fFlags,&fArgs);

}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::reset(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{  
  fFlags = 0L;
}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::setColor(
 unsigned long aPixel
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPixel==fPixel) return;
  fPixel = aPixel;
  fArgs.foreground  = aPixel;
  fFlags |= GCForeground;
}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::setLineWidth (
 int aWidth 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aWidth==fLineWidth) return;
  fLineWidth = aWidth;
  fArgs.line_width = aWidth<=1 ? 0 : aWidth;
  fFlags |= GCLineWidth;
}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::setLinePattern(
 SbLinePattern aPattern
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  static char line_style_dashed[] = {6,3};
  static char line_style_dotted[] = {3,6};
  static char line_style_dashed_dotted[] = {6,3,3,3};

  if(aPattern==fLinePattern) return;
  fLinePattern = aPattern; 
  switch(aPattern) {
  case SbLinePattern_solid:
    fArgs.line_style = LineSolid;
    fFlags |= GCLineStyle;
    break;
  case SbLinePattern_dashed:
    fArgs.line_style = LineOnOffDash;
    XSetDashes(fDisplay,fGC,0,line_style_dashed,2);
    fFlags |= GCLineStyle;
    break;
  case SbLinePattern_dotted:
    fArgs.line_style = LineOnOffDash;
    XSetDashes(fDisplay,fGC,0,line_style_dotted,2);
    fFlags |= GCLineStyle;
    break;
  case SbLinePattern_dash_dotted:
    fArgs.line_style = LineOnOffDash;
    XSetDashes(fDisplay,fGC,0,line_style_dashed_dotted,4);
    fFlags |= GCLineStyle;
    break;
  }
}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::setAreaStyle (
 SbAreaStyle aStyle 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#define stipple_hatched_width 16
#define stipple_hatched_height 16
  static unsigned char stipple_hatched_bits[] = {
   0x00, 0x7f, 0x80, 0x3f, 0xc0, 0x1f, 0xe0, 0x0f, 0xf0, 0x07, 0xf8, 0x03,
   0xfc, 0x01, 0xfe, 0x00, 0x7f, 0x00, 0x3f, 0x80, 0x1f, 0xc0, 0x0f, 0xe0,
   0x07, 0xf0, 0x03, 0xf8, 0x01, 0xfc, 0x00, 0xfe};

#define stipple_checker_width  16
#define stipple_checker_height 16
  static unsigned char stipple_checker_bits [] = {
0xff,0x00,0xff,0x00,0xff,0x00,0xff,0x00,
0xff,0x00,0xff,0x00,0xff,0x00,0xff,0x00,
0x00,0xff,0x00,0xff,0x00,0xff,0x00,0xff,
0x00,0xff,0x00,0xff,0x00,0xff,0x00,0xff};

  if(aStyle==fAreaStyle) return;
  fAreaStyle = aStyle; 
  switch(aStyle) {
  case SbAreaSolid:
    fArgs.fill_style = FillSolid;
    fFlags |= GCFillStyle;
    break;
  case SbAreaHatched:
    fArgs.fill_style = FillOpaqueStippled;
    if(fStippleHatchedPixmap==0L) {
      fStippleHatchedPixmap = 
	XCreateBitmapFromData(fDisplay,
			      XDefaultRootWindow(fDisplay),
			      (char*)stipple_hatched_bits,
			      stipple_hatched_width,
			      stipple_hatched_height);
    }
    XSetStipple(fDisplay,fGC,fStippleHatchedPixmap);
    fFlags |= GCFillStyle;
    break;
  case SbAreaChecker:
    fArgs.fill_style = FillOpaqueStippled;
    if(fStippleCheckerPixmap==0L) {
      fStippleCheckerPixmap = 
	XCreateBitmapFromData (fDisplay,
			       XDefaultRootWindow(fDisplay),
			       (char*)stipple_checker_bits,
			       stipple_checker_width,
			       stipple_checker_height);
    }
    XSetStipple(fDisplay,fGC,fStippleCheckerPixmap);
    fFlags |= GCFillStyle;
    break;
  default:
    break;
  }
}
//////////////////////////////////////////////////////////////////////////////
void SbXContext::flush(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fFlags==0L) return;
  XChangeGC(fDisplay,fGC,fFlags,&fArgs);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
float ConvertRGB_ToGrey (
 float aRed
,float aGreen
,float aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (0.30 * aRed + 0.59 * aGreen + 0.11 * aBlue);
}

#endif
