/* ----------------------------HEPVis----------------------------------------*/
/*                                                                           */
/* Class:            Twig                                                    */
/*                                                                           */
/* Description:      OIV / Motif tree structure to control visibility, etc.  */
/*                                                                           */
/* Authors:          G. Alverson, L. Taylor, T.Paul       20 Oct 1998        */
/*                                                                           */
/*---------------------------------------------------------------------------*/

// this :
#include <HEPVis/Xt/Twig.h>

#include <math.h>
#include <Inventor/SoPath.h>
#include <Inventor/SoPickedPoint.h>

#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Frame.h>
#include <Xm/ToggleBG.h>
#include <Xm/ToggleB.h>
#include <Xm/ScrolledW.h>
#include <Xm/ArrowBG.h>
#include <Xm/ScrollBar.h>
#include <Xm/DialogS.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>

                                                   
// At least for KCC we need to compell template instanciation to avoid some
// linking problem on Linux (usual template mess...).
template class std::vector<Twig*>;

// init static items in file scope
SbString* Twig::leaders = NULL;
TwigSeparator* Twig::theCurrentIvNode = NULL;


Twig::Twig (SbString name, Bool visibility, Bool cascade) : theOstream(&std::cout) {
////////////////////////////////////////////////////////////////////////////////
// 
//  Constructor
// 
////////////////////////////////////////////////////////////////////////////////

     nameOfTwig       = name; 
     ion              = visibility;      
     icascade         = cascade;
     twigWithDialog   = FALSE;

     theParentWidget  = NULL;
     showWidget       = NULL; 
     parentTwig       = NULL;

     sub_form = NULL;
     sub_show_rowcol = NULL;
     formFrame = NULL;

//   Create an OIV separator
//   -----------------------

     oiv_node = new TwigSeparator; 
       
     oiv_node->setName(nameOfTwig);
  
}

              
void Twig::destroyTree(){
////////////////////////////////////////////////////////////////////////////////
// 
//    Destroys all of the children of this Twig!
//    Also destroys TwigSeparators.
//    If you want to destroy the widgets in a popup shell (but not the shell),
//    you should use the remakeWidgetTree method.  However, take care to actually
//    destroy the associated twigs too using destroyTree.
// 
////////////////////////////////////////////////////////////////////////////////

  // Remove twig List
  // ----------------
  destroyTwig();

  // Remove TwigSeparators 
  // -------------------
  if (oiv_node->getNumChildren() >= 1)
    oiv_node->removeAllChildren();

//   // If motif widgets are attached to this twig, destroy them
//   // --------------------------------------------------------
//   if (twigWithDialog)
//     XtDestroyWidget(ShowDialogShell);

}


void Twig::destroyTwig(){
////////////////////////////////////////////////////////////////////////////////
// 
//  Destroy the twig and the List of children twigs for the twig (recursive).
// 
////////////////////////////////////////////////////////////////////////////////

  if (!twigs.empty()) {
    for (List_of_Twigs::iterator i=twigs.begin(); i != twigs.end(); i++) {
      (*i)->destroyTwig(); 
      }
    twigs.clear();
    }
}
   
int Twig::numberOfChildren() const{
////////////////////////////////////////////////////////////////////////////////
// 
//  Return the number of child twigs 
// 
////////////////////////////////////////////////////////////////////////////////

#ifdef OLD_RW

  return twigs.length();
  
#else

  return twigs.size();   
  
#endif            
}
   
Twig* Twig::getChildByIndex(int iChild) const {          
////////////////////////////////////////////////////////////////////////////////
// 
//  Return pointer to i'th child 
// 
////////////////////////////////////////////////////////////////////////////////

Twig* tempTwig = NULL;


   if (!twigs.empty() && iChild >= 0 && iChild <= (int)twigs.size()) {
       tempTwig = twigs[iChild];     
   }
   
   else{
     std::cout << "Twig::getChildByIndex>  Error for Twig = "<< nameOfTwig << ",  iChild = " << iChild << std::endl;
  }

   return tempTwig;    

}


void Twig::print(std::ostream &os) const {
////////////////////////////////////////////////////////////////////////////////
// 
//  Use with overloaded << to allow cout operations on a twig
// 
////////////////////////////////////////////////////////////////////////////////

      os << nameOfTwig.getString();

}
                  
void Twig::addChild(Twig *t) {   
////////////////////////////////////////////////////////////////////////////////
// 
//   Add a child twig (i.e. *t) to this twig (i.e. *this)
// 
////////////////////////////////////////////////////////////////////////////////

  //   tell the child twig that this twig is its parent

  twigs.push_back(t);
  t->parentTwig = this;

  //   tell this IV node about the child IV node  

  if (t->getIvNode() != NULL) {
    oiv_node->addChild(t->getIvNode());
 
  }
}
 

void Twig::printTree() {
////////////////////////////////////////////////////////////////////////////////
// 
//   Walks down the tree printing contents of twigs, sub-twigs, etc.
// 
////////////////////////////////////////////////////////////////////////////////

        leaders = new SbString("Twig >  ");
        std::cout << std::endl << std::endl;

        printMe();

        std::cout << std::endl << std::endl;
        //      delete leaders;
        leaders->makeEmpty(TRUE);
}

void Twig::printMe()  {
////////////////////////////////////////////////////////////////////////////////
// 
//   Print a twig (recursive)
// 
////////////////////////////////////////////////////////////////////////////////

     std::cout << leaders->getString() << " |---> " << getName();
     if (ion) {
       std::cout <<  " (ON)";
        }
     else{
        std::cout <<  " (OFF)";
        }
     std::cout << std::endl;

     *leaders += "       ";

    if (!twigs.empty()) {
      for (List_of_Twigs::iterator i=twigs.begin(); i != twigs.end(); i++) {
        (*i)->printMe();
      }
    }

     leaders->deleteSubString((*leaders).getLength()-7,-1);

}


void Twig::drawTree()  {
////////////////////////////////////////////////////////////////////////////////
// 
//    Walks down the tree drawing contents of twigs, sub-twigs, etc.
// 
////////////////////////////////////////////////////////////////////////////////

      drawMe();
       
      if (ion == True) {
          if (!twigs.empty()) {
            for (List_of_Twigs::iterator i=twigs.begin(); i != twigs.end(); i++) {
                (*i)->drawTree();
                }
            }
        }
}

void Twig::drawMe() {
////////////////////////////////////////////////////////////////////////////////
// 
//   Draw a twig (recursive)
//   each derived class should have its own drawMe member
// 
////////////////////////////////////////////////////////////////////////////////

//
//    Use theCurrentIvNode as "global place" where e.g. GEANT3 drawing routines can find out 
//    the IV node on which to hang their So objects (yeah - we know it's a kludge)
//

      Twig::theCurrentIvNode = oiv_node;
  
      if (ion==True && parentTwig != NULL) {
        if ((parentTwig->getIvNode())->findChild(oiv_node) < 0){
          (parentTwig->getIvNode())->addChild(oiv_node);}
      }
      else {
        if (oiv_node != NULL && parentTwig != NULL) {
          oiv_node->ref();             // don't lose the node when we unattach it
          (parentTwig->getIvNode())->removeChild(oiv_node);}
      }            
      
}

void Twig::clearTree() {
////////////////////////////////////////////////////////////////////////////////
// 
// Removes all children (oiv_nodes, not Twigs) of this oiv_node, then recreates
// the oiv_node structure to match the Twig structure
// 
////////////////////////////////////////////////////////////////////////////////

  // First attach all sub-twig seperators (not attached if invisible) 

  attachChildren();   // recursive, so also does grandchildren, etc.


  
  // Now that all sub-twig seperators are attached, blow them all away and remake tree

  if (oiv_node->getNumChildren() >= 1) {
    oiv_node->removeAllChildren();
    recreateIvTree();
  }

}

void Twig::attachChildren() {
////////////////////////////////////////////////////////////////////////////////
// 
//   Walks down the tree attaching child separators if necessary 
//   (i.e. the invisible ones which are otherwise not attached)
// 
//   Generally used only by a clearTree to make sure 
//   everything is attached prior to clearing
//
////////////////////////////////////////////////////////////////////////////////

  if (!twigs.empty()) {
    for (List_of_Twigs::iterator i=twigs.begin(); i != twigs.end(); i++) {
      
      if (!(*i)->getVisibility()) {               // child is not visible => it is not attached         
	oiv_node->addChild( (*i)->getIvNode() );  // re-attach the child
	(*i)->attachChildren();                   // attach grandchildren to child
      }   
    }
  } 
}


void Twig::recreateIvTree(){
////////////////////////////////////////////////////////////////////////////////
// 
//  Trace through the Twig tree (starting from this Twig) and create
//  oiv_nodes for each subTwig.  This does not check whether the existing
//  oiv_nodes point to anything in particular; it just makes new ones according
//  to the organization of the Twigs, so this function should not be invoked 
//  just for the hell of it.  Normally, this function will only be
//  invoked by clearTree.
// 
////////////////////////////////////////////////////////////////////////////////
  
  if (!twigs.empty()) {
    for (List_of_Twigs::iterator i=twigs.begin(); i != twigs.end(); i++) {
        (*i)->oiv_node = new TwigSeparator;
	oiv_node->addChild((*i)->oiv_node);
	(*i)->recreateIvTree();
    }
  }  
}


int Twig::createShowDialog(){
////////////////////////////////////////////////////////////////////////////////
// 
//  create show dialog widget (new show menu) 
// 
////////////////////////////////////////////////////////////////////////////////
int          n;
Arg          args[20];
Widget       ShowCloseRowcol, ShowRefreshButton, ShowCloseButton;
Widget       scroll; 
char         title[15];
/*---------------------------------------------------------------------------*/

 if (showWidget == NULL) {

   twigWithDialog = TRUE;       // flag this as the twig at top of widget menu

   strcpy(title,"");
   const char *chname = nameOfTwig.getString();
   strcat(title,chname);
   
   ShowDialogShell  = XtVaCreatePopupShell( "ShowPanel",
                      xmDialogShellWidgetClass, theParentWidget,
                      XmNtitle, title,
                      NULL);

   ShowFormDialog = XtVaCreateWidget(       "ShowForm",
                    xmFormWidgetClass,     ShowDialogShell,
                    XmNautoUnmanage,       False,
                    XmNdefaultPosition,    True,
                    XmNx,                  1000,
                    XmNy,                  740,
                    NULL);

//
// Refresh and Close buttons
// -------------------------
      n = 0;
      XtSetArg(args[n], XmNpacking,            XmPACK_COLUMN);   n++;
      XtSetArg(args[n], XmNorientation,         XmHORIZONTAL);   n++;
      XtSetArg(args[n], XmNnumColumns,                     1);   n++;
      XtSetArg(args[n], XmNadjustLast,                 False);   n++;
      XtSetArg(args[n], XmNleftWidget,         ShowFormDialog);  n++;
      XtSetArg(args[n], XmNbottomWidget,       ShowFormDialog);  n++;
      XtSetArg(args[n], XmNleftAttachment,     XmATTACH_FORM);   n++;
      XtSetArg(args[n], XmNbottomAttachment,   XmATTACH_FORM);   n++;
      
     ShowCloseRowcol = XmCreateRowColumn(ShowFormDialog,"ShowCloseRowcol",args,n);

     ShowRefreshButton= XtVaCreateManagedWidget(  " Refresh ",
                        xmPushButtonGadgetClass,  ShowCloseRowcol,
                        NULL );
                     
     ShowCloseButton  = XtVaCreateManagedWidget(  " Close ",
                        xmPushButtonGadgetClass,  ShowCloseRowcol,
                        NULL );

//
// Scroll window
// -------------
     Dimension scrollWidth, scrollHeight;
     getScrollSize(scrollWidth, scrollHeight);

   scroll = XtVaCreateWidget ("scroll",
	    xmScrolledWindowWidgetClass,  ShowFormDialog,
            XmNtopWidget,                 ShowFormDialog,
            XmNleftWidget,                ShowFormDialog,
            XmNrightWidget,               ShowFormDialog,
            XmNbottomWidget,              ShowCloseRowcol,
	    
            XmNtopAttachment,             XmATTACH_FORM,
	    XmNleftAttachment,            XmATTACH_FORM,
            XmNrightAttachment,           XmATTACH_FORM,
	    XmNbottomAttachment,          XmATTACH_WIDGET,
	    		      
            XmNwidth,                     scrollWidth,
	    XmNheight,                    scrollHeight,

            XmNvisualPolicy,    XmCONSTANT,                 // normally this is xmCONSTANT
            XmNscrollingPolicy, XmAUTOMATIC,
            NULL);

//                                                                           
//   Create the composite widget menu show (first make RowColumn container)  
//---------------------------------------------------------------------------


       n = 0;
       XtSetArg(args[n], XmNpacking,            XmPACK_TIGHT);   n++;
       XtSetArg(args[n], XmNorientation,          XmVERTICAL);   n++;
       XtSetArg(args[n], XmNnumColumns,                    3);   n++;
       XtSetArg(args[n], XmNadjustLast,                False);   n++;
 
       XtSetArg(args[n], XmNtopWidget,        scroll);           n++;
       XtSetArg(args[n], XmNleftWidget,       scroll);           n++;
       XtSetArg(args[n], XmNrightWidget,      scroll);           n++;
       XtSetArg(args[n], XmNbottomWidget,     scroll);           n++;
 
       XtSetArg(args[n], XmNtopAttachment,    XmATTACH_WIDGET);  n++;
       XtSetArg(args[n], XmNleftAttachment,   XmATTACH_WIDGET);  n++;
       XtSetArg(args[n], XmNrightAttachment,  XmATTACH_WIDGET);  n++;
       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET);  n++;
   
       ShowRowcol = XmCreateRowColumn(scroll,"ShowRowcol",args,n);

//         
// Add the buttons for all the subtwigs
// ------------------------------------

     showCreateItem(ShowRowcol);

 
//                                                                           
//   Add Callbacks for bottom buttons                                                                            
//------------------------------------
 
     XtAddCallback(ShowRefreshButton, XmNactivateCallback, (XtCallbackProc) showRefreshCB,
                                      (XtPointer) this);     
     XtAddCallback(ShowCloseButton,   XmNactivateCallback, (XtCallbackProc) showCloseCB,
                                      (XtPointer) this);
     showWidget = ShowFormDialog;
 
//
// Manage all the widgets
// ----------------------

     XtManageChild (ShowCloseRowcol);
     XtManageChild (ShowRowcol);
     XtManageChild (scroll);

  } 
     XtManageChild   (showWidget);
     XmUpdateDisplay (showWidget);                    

     return 0;     
}
 
void Twig::showCreateItem(Widget parentWidget){ 
////////////////////////////////////////////////////////////////////////////////
// 
//  Create part of a show widget comprising a toggle button, label, and 
//  possibly a row-column to hold further items
// 
////////////////////////////////////////////////////////////////////////////////
  Arg      args[20];
  int      n;
  XmString label;


  formFrame = XtVaCreateManagedWidget("ShowItems", xmFrameWidgetClass,
              parentWidget, XmNshadowType, XmSHADOW_ETCHED_IN, NULL);


  if (twigs.empty()){
    
    // No children => make button+label only (and no cascade toggle)
    // -------------------------------------------------------------
    
    n = 0;
    char chname[100];
    strcpy(chname,nameOfTwig.getString());
    label = XmStringCreateLocalized(chname);
    XtSetArg(args[n],XmNlabelString, label);                 n++;
    
    if (ion) {
      XtSetArg(args[n], XmNset, True);                       n++;}
    else {
      XtSetArg(args[n], XmNset, False);                      n++;}
    
    XtSetArg(args[n],XmNalignment,   XmALIGNMENT_BEGINNING); n++;
     
    toggleButton = XmCreateToggleButton(formFrame,"button",args,n);
    XmStringFree(label);
    XtManageChild(toggleButton);
    XtManageChild(formFrame);
    XtAddCallback(toggleButton, XmNvalueChangedCallback, (XtCallbackProc) showCB,
		   (XtPointer) this);
  }

  else {         


    // Children => make button+label and (recursively) a form for children to go in
    // ----------------------------------------------------------------------------

     
    sub_form = XtVaCreateWidget("ShowSubForm",
          			 xmFormWidgetClass,formFrame,
				 NULL);
     
    // create toggle button + label 
    // ----------------------------
     
    n = 0;
    XtSetArg(args[n], XmNtopWidget,        sub_form);        n++;
    XtSetArg(args[n], XmNleftWidget,       sub_form);        n++;
    
    XtSetArg(args[n], XmNtopAttachment,    XmATTACH_FORM);   n++;
    XtSetArg(args[n], XmNleftAttachment,   XmATTACH_FORM);   n++;
     
    char chname[100];
    strcpy(chname,nameOfTwig.getString());
    label = XmStringCreateLocalized(chname);
    XtSetArg(args[n], XmNlabelString, label);                n++;

    
    // Set initial state of toggle button (on or off)
    // ----------------------------------------------    
 
    if (ion) {
      XtSetArg(args[n], XmNset, True);                       n++;
    }
    else {
      XtSetArg(args[n], XmNset, False);                      n++;
    }
     
    XtSetArg(args[n], XmNalignment,  XmALIGNMENT_BEGINNING); n++;

    toggleButton = XmCreateToggleButton(sub_form,"button",args,n);

    
    // Make the cascade-inducing pushbutton and determine 
    // which way the arrow on it should face.
    //----------------------------------------------------

    n = 0;
    XtSetArg(args[n], XmNtopWidget,        sub_form);        n++;
    XtSetArg(args[n], XmNleftWidget,       toggleButton);    n++;
    XtSetArg(args[n], XmNtopAttachment,    XmATTACH_FORM);   n++;
    XtSetArg(args[n], XmNleftAttachment,   XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNlabelType,        XmPIXMAP);        n++;
    if (icascade) {
      XtSetArg(args[n], XmNarrowDirection, XmARROW_LEFT   ); n++;
    }
    else {
      XtSetArg(args[n], XmNarrowDirection, XmARROW_RIGHT  ); n++;
    }
   
    cascadeButton = XmCreateArrowButtonGadget(sub_form,"",args,n);

    XmStringFree(label);
     
    XtAddCallback(toggleButton, XmNvalueChangedCallback, (XtCallbackProc) showCB,
                  (XtPointer) this);
   
    XtAddCallback(cascadeButton, XmNactivateCallback, (XtCallbackProc) cascadeCB,
                  (XtPointer) this);


    // create a row-column to the right of the buttons 
    // -----------------------------------------------
     
    n = 0;
    XtSetArg(args[n], XmNpacking,           XmPACK_TIGHT);    n++;
    XtSetArg(args[n], XmNorientation,          XmVERTICAL);   n++;
    XtSetArg(args[n], XmNnumColumns,                    1);   n++;
    XtSetArg(args[n], XmNadjustLast,                False);   n++;
    
    XtSetArg(args[n], XmNtopWidget,        sub_form);         n++;
    XtSetArg(args[n], XmNleftWidget,       cascadeButton);    n++;
    XtSetArg(args[n], XmNrightWidget,      sub_form);         n++;
    XtSetArg(args[n], XmNbottomWidget,     sub_form);         n++;
    
    XtSetArg(args[n], XmNtopAttachment,    XmATTACH_FORM);    n++;
    XtSetArg(args[n], XmNleftAttachment,   XmATTACH_WIDGET);  n++;
    XtSetArg(args[n], XmNrightAttachment,  XmATTACH_FORM);    n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);    n++;
     
    sub_show_rowcol = XmCreateRowColumn(sub_form,"sub_ShowRowcol",args,n);

    // now recursively do the same as above for sub-twigs
    // --------------------------------------------------
     
    if ( this->icascade ) {
      for (List_of_Twigs::iterator i=twigs.begin(); i != twigs.end(); i++) {
        (*i)->showCreateItem(sub_show_rowcol); }
     }
     
     XtManageChild(sub_show_rowcol);       
     XtManageChild(toggleButton);     
     XtManageChild(cascadeButton);     
     
     XtManageChild(sub_form);
     XtManageChild(formFrame);     
   }
 }


void Twig::getScrollSize(Dimension& width, Dimension& height) {
////////////////////////////////////////////////////////////////////////////////
// 
//  Make a dummy arrangement of everything that will go inside the scroll
//  widget created in createShowDialog, then check its size.  This will 
//  allow the automatic scroll widget to be initialized to the size of the
//  original Twig tree.  This is arguably lame, but the alternative is 
//  to set up application-defined scrolling.
// 
////////////////////////////////////////////////////////////////////////////////
int        n;
Arg        args[20];
Widget     bogusForm, bogusRowcol, bogusScrollBar, bogusShell; 


 bogusShell = XtVaCreatePopupShell( "bogusshell",
                xmDialogShellWidgetClass, theParentWidget,
                NULL);

 bogusForm  = XtVaCreateWidget(      "bogusForm",
                xmFormWidgetClass,     bogusShell,
                NULL);

  n = 0;
  XtSetArg(args[n], XmNpacking,            XmPACK_TIGHT);   n++;
  XtSetArg(args[n], XmNorientation,          XmVERTICAL);   n++;
  XtSetArg(args[n], XmNnumColumns,                    3);   n++;
  XtSetArg(args[n], XmNadjustLast,                False);   n++;
  
  XtSetArg(args[n], XmNtopWidget,        bogusForm);        n++;
  XtSetArg(args[n], XmNleftWidget,       bogusForm);        n++;
  XtSetArg(args[n], XmNrightWidget,      bogusForm);        n++;
  XtSetArg(args[n], XmNbottomWidget,     bogusForm);        n++;

  XtSetArg(args[n], XmNtopAttachment,    XmATTACH_FORM);    n++;
  XtSetArg(args[n], XmNleftAttachment,   XmATTACH_FORM);    n++;
  XtSetArg(args[n], XmNrightAttachment,  XmATTACH_FORM);    n++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);    n++;
  
  bogusRowcol = XmCreateRowColumn(bogusForm,"bogusRowcol",args,n);
  
  showCreateItem(bogusRowcol);

  // this shrinks the gui for some reason...  
  XtManageChild (bogusRowcol);

  XtManageChild (bogusForm);

  // Size of the stuff that goes in the scroll widget
  //------------------------------------------------

   XtVaGetValues(bogusForm,
        	XmNwidth,  &width,
    	        XmNheight, &height, NULL);


  // Check how big a scroll bar is (turns out to be 15)
  // --------------------------------------------------

  bogusScrollBar = XtVaCreateManagedWidget("bogusScrollBar",
                   xmScrollBarWidgetClass, bogusForm,
                   XmNorientation, XmVERTICAL,					     
		   NULL);

  Dimension scrollBarWidth;
  XtVaGetValues(bogusScrollBar,
        XmNwidth, &scrollBarWidth,
	NULL);

  // Final dimensions for the scroll widget
  //---------------------------------------

  width  += scrollBarWidth;
  height += scrollBarWidth;

  XtDestroyWidget(bogusShell);

  return;
}


void Twig::showCB(Widget, XtPointer client_data,XtPointer call_data){
////////////////////////////////////////////////////////////////////////////////
// 
//  Invoked when a visibility toggle (see: Twig::createShowDialog()) is pressed 
// 
////////////////////////////////////////////////////////////////////////////////
    Twig *t = (Twig *) client_data;
    XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) call_data;

    if (cbs->set) {
      t->ion = True;
      }
    else {
      t->ion = False; 
      }         
}        

void Twig::cascadeCB(Widget, XtPointer client_data,XtPointer){
////////////////////////////////////////////////////////////////////////////////
// 
//  Invoked when a cascade button is pressed.
//  Toggles the value of icascade and remakes widget accordingly.
// 
////////////////////////////////////////////////////////////////////////////////
  Twig *t = (Twig *) client_data;
  
  if (t->icascade) {
    t->setCascade(FALSE);
  }
  else {
    t->setCascade(TRUE);
  }
  
  Twig* currentTwig = t;
  for (int i=0 ; i<100 ; i++) {
    if (currentTwig->twigWithDialog)
      {
	break;
      }
    if (currentTwig->parentTwig != NULL)
      {
	currentTwig = currentTwig->parentTwig;
      }
    else
      {
	std::cout << "** Twig Warning in cascadeCB. " << std::endl;
	std::cout << "Arrived at the top of the twig tree " << std::endl;
	std::cout << "but didn't find a TRUE twigWithDialog flag." << std::endl;
	std::cout << "Therefore, widget will not be destroyed." << std::endl;
	return;
      }
  }
     
  if (currentTwig->formFrame != NULL ) {
    XtDestroyWidget(currentTwig->formFrame);
    currentTwig->formFrame = NULL;
  }

  currentTwig->showCreateItem(currentTwig->ShowRowcol);

}


void Twig::remakeWidgetTree(Twig* t){
////////////////////////////////////////////////////////////////////////////////
// 
//  Find twig that did the createShowDialog, and destroy the Motif stuff
//  with the exception of the popup shell.  This is used in anticipation of
//  the destruction of a twig tree and creation of a new twig tree that should
//  display its contents in the same popup shell. 
// 
////////////////////////////////////////////////////////////////////////////////

    
  Twig* currentTwig = t;
  for (int i=0 ; i<100 ; i++) {
    if (currentTwig->twigWithDialog)
      {
	break;
      }
    if (currentTwig->parentTwig != NULL)
      {
	currentTwig = currentTwig->parentTwig;
      }
    else
      {
	std::cout << "** Twig Warning in clearWidgetTree. " << std::endl;
	std::cout << "Arrived at the top of the twig tree " << std::endl;
	std::cout << "but didn't find a TRUE twigWithDialog flag." << std::endl;
	std::cout << "Therefore, widget will not be destroyed." << std::endl;
	return;
      }
  }
     
  if (currentTwig->formFrame != NULL ) {
    XtDestroyWidget(currentTwig->formFrame);
    currentTwig->formFrame = NULL;
  }

  currentTwig->showCreateItem(currentTwig->ShowRowcol);

}
///////////////////////////////////////////////
//
//  Picking
//
///////////////////////////////////////////////
const SoPickedPoint* Twig::getCurrentPick() const {
  return pickedPoint;
}
void Twig::setCurrentPick(const SoPickedPoint *pick) 
{
  pickedPoint = pick;
}
SoPath*  Twig::pickCB (void *theCBStruct, const SoPickedPoint *pick)
{
////////////////////////////////////////////////////////////////////////////////
//
//  pickCB: A pick as opposed to a selection (see selectionCB) is used
//          when either (1) more information than just the path to the object
//          is required, eg, the SoPointDetail info, or (2) the path needs
//          to be edited before the selection callback is fired.
//
////////////////////////////////////////////////////////////////////////////////

// default behavior is to save the pick information and fire the selectionCB 
// routines

// Pull the items from the Callback structure
   Twig *twig = ((TwigCallbackStruct*) theCBStruct)->theTwig;

  twig->setCurrentPick(pick);
  return pick->getPath();
}  
void Twig::selectionCB  (void *theCBStruct, SoPath *selectionPath) {
////////////////////////////////////////////////////////////////////////////////
//
//  selectionCB: CB routine for OIV pick.
//               Searches for a Twig to match this OIV node and lets it 
//               do the work.
// 
////////////////////////////////////////////////////////////////////////////////
 

// Pull the items from the Callback structure
   Twig *twig = ((TwigCallbackStruct*) theCBStruct)->theTwig;
   
// work up the path looking for an SoSeparator which is the 
// separator matching a Twig   
   SoNode *bottomNode = selectionPath->getTail();
   const int pathNumberNodes = selectionPath->getLength();
   
   int pathIndex = 0;
   bool found = false; Twig *matchTwig = NULL;   
   while ( (pathIndex < pathNumberNodes) && !found ){ // double check this line 
     SoNode *aNode = selectionPath->getNodeFromTail(pathIndex);
//   jump over any node not a separator   
     while (!aNode->isOfType(SoSeparator::getClassTypeId()) && pathIndex < pathNumberNodes) {
       aNode = selectionPath->getNodeFromTail(++pathIndex);
       }
//   have a separator; look for it in the Twig tree 
     if ((matchTwig = twig->findTwigNodeMatch(aNode)) != NULL) {
       found = true; }
     else {
       pathIndex++; }               
   }
   if (found) {
     *(twig->theOstream) << "Found a match in tree "<< twig->getName() << ": " << std::endl;
     matchTwig->setCurrentPick(twig->getCurrentPick()); // set pick info if present
     matchTwig->printInfo(*(twig->theOstream), bottomNode);
     *(twig->theOstream) << std::endl;}
   else {
     *(twig->theOstream) << "No match for selected object!" << std::endl;}
   }
Twig* Twig::findTwigNodeMatch(const SoNode *node) {
////////////////////////////////////////////////////////////////////////////////
// 
//  findTwigNodeMatch: search for a particular SoNode in the Twig tree
//  returns pointer to same or null. 
//
////////////////////////////////////////////////////////////////////////////////
 
  Twig *found_Twig = NULL;
//  Is this node it? Note that we're only looking for a match for TwigSeparators,
//  not just any separator, so we don't have to look down from the oiv_node.
//TR  More optional trace below
//  SoType nodeType = oiv_node->getTypeId();
//  if (!nodeType.isBad()) {
//    SbName nodeName = nodeType.getName();
//    std::cout << ", a "<<nodeName.getString()<<std::endl;
//    }
//  else {
//    std::cout << std::endl; 
//    }
  
  if (oiv_node == node) {found_Twig = this;}
  else {
    bool found = false;
    if (!twigs.empty()) {
      for (List_of_Twigs::iterator i=twigs.begin(); (i != twigs.end()) && !found ; i++) {
        found_Twig = (*i)->findTwigNodeMatch(node);
	found = (found_Twig != NULL); }
    }	
  }  
  return found_Twig;
}  	           
void Twig::printInfo(std::ostream& ostr, const SoNode*) const {
////////////////////////////////////////////////////////////////////////////////
//
//  printInfo: prints any information available concerning an OIV node attached
//  to the Twig.
//
////////////////////////////////////////////////////////////////////////////////

  ostr << "Twig " << getName() <<": ";
  ostr << "No additional information available on this node." <<std::endl;
}
Bool Twig::okToRedraw() {

//
// In most instances this is the basic test to see if a node can
// and should be drawn. The criteria are
// 1) The visibility flag must be set on
// 2) The oiv node must exist, otherwise the Twig::drawMe() failed
// 3) The node shouldn't have pre-existing children

   return (ion==TRUE && oiv_node != NULL && oiv_node->getNumChildren() == 0);
}          
void Twig::setOstream(std::ostream& os) {
// We need some place to keep track of where we ship output
   theOstream = &os;}
     
       
