//  Mai/2007 : G.Barrand : creation.

// The scene is a grey floor with a red cube over it.
// At startup, the camera is slightly on top of the floor
// looking at the cube.

#include <Inventor/SDL/SoSDL.h>
#include <Inventor/SDL/viewers/SoSDL_Viewer.h>

#include <Inventor/SDL/app.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDirectionalLight.h>

#ifdef SDL_WILD
#include <Lib/InternetSocket.h>
#include <Inventor/actions/SoWriteAction.h>
#include "camera.h"
#include "net.h"
#include <iostream>

static SbBool curl_file(const SbString& a_url,const SbString& a_out_file,void* a_tag) {
  WILD_Tag* WILD_tag = (WILD_Tag*)a_tag;
  Lib::BaseSocket* socket = WILD_tag->m_socket;

  //::printf("debug : SoSDL_web::curl_file : %d url \"%s\"...\n",
  //   getpid(),a_url.getString());

 {SDL_Event event;
  event.user.code = WILD_USEREVENT_CURL;
  fill_SDL_Event_with_string(a_url.getString(),event);
  if(!send_SDL_Event(*socket,event)) {
    std::cout << "SoSDL_web::curl_file : send_uchar failed." << std::endl;
    delete [] ((char*)event.user.data2);
    return FALSE;
  }
  delete [] ((char*)event.user.data2);}

  //printf("debug : SoSDL_web::curl_file : wait_file : %d port %d url %s : begin\n",getpid(),socket->port(),a_url.getString());

  while(true) { //wait a WILD_USEREVENT_FILE.
  bool have_input;
  if(!socket->is_there_input(have_input)) return FALSE;
  if(have_input) {
    SDL_Event event;
    if(!wait_SDL_Event(*socket,event)) {
      std::cout << "SoSDL_web::curl_file : wait_SDL_Event failed." 
                << std::endl;
      return FALSE;
    }
    if((event.type==SDL_USEREVENT)&&(event.user.code==WILD_USEREVENT_FILE)){
      std::string out_file(a_out_file.getString());
      unsigned int length;
      if(!read_SDL_Event_file(event,out_file,length)) return FALSE;
      if(length==0) {
        //NOTE : SoFile stops on an empty file !
        std::cout << "SoSDL_web::curl_file :"
                  << " WARNING : receive an empty file."
                  << std::endl;
        return FALSE;
      }
      break;
    } else {
      SDL_Event* evt = new SDL_Event;
      ::memcpy(evt,&event,sizeof(SDL_Event));
      WILD_tag->m_events.push_back(evt);
    }
  }}

  //printf("debug : SoSDL_web::curl_file : wait_file : %d port %d url %s : end\n",getpid(),socket->port(),a_url.getString());

  return TRUE;
}
#else
#include <math.h>
#ifndef M_PI
#define M_PI       3.1415926535897931160E0
#define M_PI_2     1.5707963267948965580E0  
#endif
#define FM_PI      ((float)M_PI)
#define FM_PI_2    ((float)M_PI_2)
#include <Inventor/nodes/SoPerspectiveCamera.h>
#endif

//Windows, Darwin : "main" becomes "SDL_main" through a cpp macro in SDL.h.

#include <Lib/SoMain.h>
#include <Lib/SoCurl.h>

//#define BRING_HEPVIS
#ifdef BRING_HEPVIS
#include <HEPVis/fields/SoSFPolyhedron.h>
#include <HEPVis/actions/SoAlternateRepAction.h>
#include <HEPVis/actions/SoGL2PSAction.h>
#include <HEPVis/nodes/SoPolyhedron.h>
#include <HEPVis/nodes/SoHighlightMaterial.h>
#include <HEPVis/nodes/SoMarkerSet.h>
#include <HEPVis/nodekits/SoDetectorTreeKit.h>
#endif

#include "SoTools.h"

#define HAS_MENU

#define SOSDL_WEB

#ifdef HAS_MENU
#include "../applications/menu.h"
#endif

//
#define HAS_SHOOT

#ifdef HAS_SHOOT
#include "../applications/shoot.h"
#endif

#define HAS_SELECTION

#ifdef HAS_SELECTION
#include "../applications/selection.h"
#endif

#include <Slash/Tools/args.h>
#include <Slash/Tools/system.h>
#include <Slash/Tools/file.h>
#include <Slash/Tools/fmanip.h>

#include <cstdio>
#include <iostream>
#include <cstdlib>

int main(int argc,char* argv[]) {

  bool is_mac_app = false;
#ifdef __APPLE__
  //NOTE : we can't test for argv[1] with -psn_ since
  //       SDLMain/main strips the argv[1] argument.
  std::string contents_dir;
  if(argc>0) {
    std::string prog_dir = Slash::dir_name(argv[0]);
    if(Slash::base_name(prog_dir)=="MacOS") {
      is_mac_app = true;
      contents_dir = Slash::dir_name(prog_dir);
    }
  }
#endif

  ///////////////////////////////////////////////
  // Init :
  ///////////////////////////////////////////////
  Slash::args args(argc,argv);

  int verbose = 0;
  if(is_mac_app) { 
    if(!Slash::get_env<int>("SOSDL_VERBOSE",verbose)) verbose = 0;
  } else {
    if(args.is_arg("-verbose")) verbose = 1;
  }

  bool full_screen = true; //default
  if(is_mac_app) { 
    if(::getenv("SOSDL_NOT_FULL_SCREEN")) full_screen = false;
  } else {
    if(args.is_arg("-not_full_screen")) full_screen = false;
  }

#define DEF_URL \
  "http://OpenScientist.lal.in2p3.fr/download/data/iv/so_web.iv";

  std::string URL;
  if(is_mac_app) {
    if(!Slash::getenv("SOSDL_URL",URL)) URL = DEF_URL;
  } else {
    if(!args.find("-url",URL)) URL = DEF_URL;
  }
  if(verbose) {
    std::cout << "URL is \"" << URL << "\"." << std::endl;
  }

#ifdef SDL_WILD
  bool local_curl = false;
  if(is_mac_app) { 
    if(Slash::isenv("SOSDL_LOCAL_CURL")) local_curl = true;
  } else {
    if(args.is_arg("-local_curl")) local_curl = true;
  }
#endif

  SoSDL_Widget* topWidget = 0;
  if(full_screen) {
    topWidget = SoSDL::init(0,0,SDL_FULLSCREEN);
  } else {
    float aspect = 1680.0F/1050.0F;
    unsigned int ww = 800;
    unsigned int wh = (((float)ww)/aspect); //inside window height
    topWidget = SoSDL::init(ww,wh,0);
  }
  if(!topWidget) {
    std::cout << "SoSDL_web : SoSDL::init failed." << std::endl;
    return EXIT_FAILURE;
  }

  SDL_Surface* screen = SoSDL::screen();
  if(!screen) {
    std::cout << "SoSDL_web : SoSDL::sceen() return 0." << std::endl;
    return EXIT_FAILURE;
  }

  //printf("debug : surf %d %d\n",screen->w,screen->h);
  unsigned int ww = screen->w;
  unsigned int wh = screen->h;
  float aspect = ((float)ww)/((float)wh);

#ifdef SDL_WILD
#include "../applications/WILD_start.icc"
#else
  SDL_ShowCursor(1);
#endif

#ifdef HAS_SHOOT
#ifdef SDL_WILD
  //arrange to have the mire not on a screen separator.
  float shoot_fact_x = (2*int(screen_cols/2)==screen_cols)?0.55F:0.5F;
  float shoot_fact_y = (2*int(screen_rows/2)==screen_rows)?0.55F:0.5F;
  int shoot_x = WILD_ww*shoot_fact_x;
  int shoot_y = WILD_wh*shoot_fact_y;
  cvt_WILD_to_local(screen_col,screen_row,ww,wh,shoot_x,shoot_y);
#else
  float shoot_fact_x = 0.5F;
  float shoot_fact_y = 0.5F;
  int shoot_x = ww*shoot_fact_x;
  int shoot_y = wh*shoot_fact_y;
#endif
  Shoot* shoot = new Shoot;
  shoot->f_shoot_x = shoot_x;
  shoot->f_shoot_y = wh-shoot_y;  // 0 = top
#endif

  SoMain::initClass();
#ifdef SDL_WILD
  if(!local_curl) SoCurl::setCurlFunction(curl_file,&WILD_tag);
#endif
  SoCurl::initClass();

#ifdef BRING_HEPVIS
  SoAlternateRepAction::initClass();
  SoGL2PSAction::initClass();
  SoSFPolyhedron::initClass();
  SoPolyhedron::initClass();
  SoHighlightMaterial::initClass();
  HEPVis_SoMarkerSet::initClass();
  SoDetectorTreeKit::initClass();
#else
#if defined(__APPLE__)
  void* HEPVis_handle = 0;
  if(is_mac_app) {
    std::string plugin = contents_dir+"/bin/HEPVisLoad.bundle";
    if(verbose) {
      std::cout << "SoSDL_web :"
                << " plugin is \"" << plugin << "\"."
                << std::endl;
    }
    HEPVis_handle = SDL_LoadObject(plugin.c_str());
  } else {
    HEPVis_handle = SDL_LoadObject("HEPVisLoad.bundle");
  }
#elif defined(WIN32)
  void* HEPVis_handle = SDL_LoadObject("HEPVisLoad.dll");
#else
  void* HEPVis_handle = SDL_LoadObject("libHEPVisLoad.so");
#endif
  if(!HEPVis_handle) {
    if(verbose) std::cout << "HEPVisLoad plugin not found." << std::endl;
  } else {
    typedef void(*Func)();
    Func func = (Func)SDL_LoadFunction(HEPVis_handle,"HEPVis_initClasses");
    if(!func) {
      if(verbose) std::cout << "HEPVis_initClasses not found." << std::endl;
    } else {
      func();
    }
  }
#endif

  SoSeparator* root = new SoSeparator;
  root->ref();

#ifdef HAS_MENU
#include "../applications/menu.ic"
#endif

#ifdef HAS_SHOOT
#include "../applications/shoot.ic"
#endif

#ifdef HAS_SELECTION
  MySelection* sceneGraph = new MySelection;
  sceneGraph->setPickRoot(sceneGraph);
  sceneGraph->addSelectionCallback(selectionCB,(void*)sceneGraph);
  sceneGraph->addDeselectionCallback(deselectionCB,(void*)sceneGraph);
  sceneGraph->policy = SoSelection::SINGLE;
  sceneGraph->policy = SoSelection::TOGGLE;
  sceneGraph->policy = SoSelection::SHIFT;
#else
  SoSeparator* sceneGraph = new SoSeparator;
#endif
  root->addChild(sceneGraph);

  JoyTag joyTag;
  add_CBs(sceneGraph,&joyTag);

  // parameters of the session :
  // the below are the defaults in SoSDL.cpp and SoSDL_viewer.cpp.
  unsigned int arg_keyRepeatDelay = 200;
  unsigned int arg_keyRepeatInterval = 10;

  float arg_deltaTranslation = 1.0F;
  float arg_deltaRotation = 3.1416F/180.0F;
  //unsigned int arg_moveTimerInterval = 30;

  SbBool arg_checkFloor = FALSE;
  float arg_floor = 0.0F;

  SbVec3f arg_camera_position(0,0,1);
  float arg_camera_nearDistance = 1;
  float arg_camera_farDistance = 10;
  float arg_camera_focalDistance = 5;

  SbBool arg_sceneRenderCaching = FALSE;

  SoSeparator* sep = 0;
 {SbString file("scene.iv");
#ifdef __APPLE__
  if(is_mac_app) {
    file = SbString(contents_dir.c_str());
    file += "/Resources/scene.iv";
  }
#endif
  if(Slash::file::exists(file.getString())==TRUE){
    SoInput soInput;
    if(soInput.openFile(file.getString())==FALSE) {
      std::cout << "SoSDL_web :"
                << " SoInput.openFile failed for " 
                << std::string(file.getString()) << "."
                << std::endl;
      return EXIT_FAILURE;
    }
    sep = SoDB::readAll(&soInput);
    if(sep) sep->ref();
  }}
  if(!sep) {
    if(URL.size()) {
      if( (URL.find("LHCb")!=std::string::npos) &&
          (SoType::fromName("SoPolyhedron")==SoType::badType()) ){
        std::cout << "SoSDL_web :"
                  << " can't load \"" << URL << "\""
                  << " because HEPVis is not loaded."
                  << std::endl;
        //not fatal
      } else {
        SoCurl* soCurl = new SoCurl;
        soCurl->ref();
        soCurl->name.setValue(URL.c_str());
        SoGroup* g = soCurl->copyChildren(); //g ref count is 0.
        if(g) {
          sep = new SoSeparator;
          sep->ref();
          sep->addChild(g);
        }
        soCurl->unref();
      }
    }
  }

  if(!sep) {
    sep = new SoSeparator;
    sep->ref();
  }

  //look for a SoMain :
  SoMain* soMain = (SoMain*)SoTools_searchNode(sep,SoMain::getClassTypeId());
  if(soMain) {
    arg_keyRepeatDelay = soMain->keyRepeatDelay.getValue();
    arg_keyRepeatInterval = soMain->keyRepeatInterval.getValue();

    arg_deltaTranslation = soMain->deltaTranslation.getValue();
    arg_deltaRotation = soMain->deltaRotation.getValue();
    //arg_moveTimerInterval = soMain->moveTimerInterval.getValue();

    arg_checkFloor = soMain->checkFloor.getValue();       
    arg_floor = soMain->floor.getValue();
    //printf("debug : SoMain : %d %g\n",arg_checkFloor,arg_floor);

    arg_camera_position = soMain->camera_position.getValue();
    arg_camera_nearDistance = soMain->camera_nearDistance.getValue();
    arg_camera_farDistance = soMain->camera_farDistance.getValue();
    arg_camera_focalDistance = soMain->camera_focalDistance.getValue();

    arg_sceneRenderCaching = soMain->sceneRenderCaching.getValue();
  }

  ///////////////////////////////////////////////
  // Camera (us) :
  ///////////////////////////////////////////////
  // Head light :
 {SoDirectionalLight* soLight = new SoDirectionalLight;
  soLight->color.setValue(1,1,1);
  // Same as Coin<Gui>/So<Gui>Viewer :
  soLight->direction.setValue(SbVec3f(1,-1,-10));
  sceneGraph->addChild(soLight);}

#ifdef SDL_WILD
  SoCamera* soCamera = create_camera(screen_cols,screen_rows,
  		                     screen_col,screen_row,
                                     aspect,2,arg_camera_nearDistance);
#else
  SoPerspectiveCamera* soCamera = new SoPerspectiveCamera;
#endif
  soCamera->position.setValue(arg_camera_position);
  soCamera->focalDistance.setValue(arg_camera_focalDistance);
  soCamera->farDistance.setValue(arg_camera_farDistance);
  sceneGraph->addChild(soCamera);

  if(arg_sceneRenderCaching==TRUE) {
    sep->renderCaching.setValue(SoSeparator::ON);
  }

  sceneGraph->addChild(sep);
  sep->unref();

  // SDL :
  SDL_EnableKeyRepeat(arg_keyRepeatDelay,arg_keyRepeatInterval);
  //SoSDL::setMoveTimerInterval(arg_moveTimerInterval);

  // Viewer :
  SoSDL_Viewer* viewer = new SoSDL_Viewer(topWidget);
#ifdef HAS_SELECTION
  viewer->setGLRenderAction(new SoBoxHighlightRenderAction());
  //viewer->setGLRenderAction(new SoLineHighlightRenderAction());
#endif
  viewer->setSceneGraph(root);
  viewer->setCamera(soCamera);
  viewer->setHeadlight(FALSE);
  viewer->setDefaultDeltaTranslation(1);
  viewer->setBackgroundColor(SbColor(1,1,1));

  //SoSDL :
  viewer->setDefaultDeltaTranslation(arg_deltaTranslation);
  viewer->setDefaultDeltaRotation(arg_deltaRotation);
  viewer->checkFloor(arg_checkFloor);
  viewer->setFloor(arg_floor); //should take into account the camera height.
  viewer->saveCamera();

  viewer->show();
    
  SoSDL::show(topWidget);

  SoSDL::mainLoop();
    
  sceneGraph->unref();

  delete viewer;

#ifdef BRING_HEPVIS
#else
  if(HEPVis_handle) {
    SDL_UnloadObject(HEPVis_handle);
  }
#endif

  SoSDL::clear();  
	
#ifdef SDL_WILD
  if(WILD_tag.m_events.size()) {
    std::cout << "SoSDL_web: WARNING :" 
              << " " << (int)WILD_tag.m_events.size() << " not processed."
              << std::endl;
  }
  delete socket;
#endif

  if(verbose) std::cout << "SoSDL viewer : exiting..." << std::endl;

  return EXIT_SUCCESS;
}
