#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoOrthographicCamera.h>

#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
//#include <Inventor/nodes/SoDirectionalLight.h>

#include <iostream>

class Menu : public SoSeparator {
public:
  Menu()
  :fIndex(0),fSwitch(0),fButtons(0)
#ifdef SDL_WILD
  ,fSocket(0)
#endif
  {}
public:
  int fIndex;
  SoSwitch* fSwitch;
  SoSeparator* fButtons;
#ifdef SDL_WILD
  Lib::BaseSocket* fSocket;
#endif
};

static SbColor bg_color() {return SbColor(1,1,1);}
static SbColor hg_color() {return SbColor(0.6F,0.6F,0.6F);}

#define CHILD_TRANS 0
#define CHILD_BACK 2
#define CHILD_MAT 3

static void next_item(Menu& aMenu){
  SoSeparator* buttons = aMenu.fButtons;

 {// reset current button :
  SoSeparator* button = (SoSeparator*)buttons->getChild(aMenu.fIndex);
  SoMaterial* material = (SoMaterial*)button->getChild(CHILD_MAT);
  material->diffuseColor.setValue(bg_color());}

  aMenu.fIndex++;
  if(aMenu.fIndex>=buttons->getNumChildren()) aMenu.fIndex = 0;
      
 {// highlight new current button :
  SoSeparator* button = (SoSeparator*)buttons->getChild(aMenu.fIndex);
  SoMaterial* material = (SoMaterial*)button->getChild(CHILD_MAT);
  material->diffuseColor.setValue(hg_color());}

}

static void prev_item(Menu& aMenu){
  SoSeparator* buttons = aMenu.fButtons;

 {// reset current button :
  SoSeparator* button = (SoSeparator*)buttons->getChild(aMenu.fIndex);
  SoMaterial* material = (SoMaterial*)button->getChild(CHILD_MAT);
  material->diffuseColor.setValue(bg_color());}

  aMenu.fIndex--;
  if(aMenu.fIndex<0) aMenu.fIndex = buttons->getNumChildren()-1;
      
 {// highlight new current button :
  SoSeparator* button = (SoSeparator*)buttons->getChild(aMenu.fIndex);
  SoMaterial* material = (SoMaterial*)button->getChild(CHILD_MAT);
  material->diffuseColor.setValue(hg_color());}

}

#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>

static void menu_keyboard_CB(void* aTag,SoEventCallback* aEventCB){
  if(aEventCB->isHandled()) return;
  const SoKeyboardEvent* event = (SoKeyboardEvent*)aEventCB->getEvent();
  if(SO_KEY_PRESS_EVENT(event,M)) {
    //printf("debug : M typed\n");
    Menu* menu = (Menu*)aTag;    
    SoSwitch* sw = menu->fSwitch;    
    sw->whichChild.setValue(sw->whichChild.getValue()?0:1);
    aEventCB->setHandled();

  } else if(SO_KEY_PRESS_EVENT(event,RETURN)) {
    Menu* menu = (Menu*)aTag;    
    SoSwitch* sw = menu->fSwitch;    
    if(sw->whichChild.getValue()==1) {
      //NOTE : WILD : if menu items are globals to the application,
      //              only one screen should do something ! 
      //actions :
      if(menu->fIndex==0) { //Exit //global
#ifdef SDL_WILD
        SDL_Event event;
        event.type = SDL_QUIT;
        if(!send_SDL_Event(*(menu->fSocket),event)) {/*FIXME*/}
#endif
        SoSDL::toQuit();
        aEventCB->setHandled();

      } if(menu->fIndex==1) { //Home //local
        SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
        viewer->restoreCamera();
        aEventCB->setHandled();

#ifdef SOSDL_WEB
#define STATIC_SCENE "staticScene"
#define DYNAMIC_SCENE "dynamicScene"
      } if(menu->fIndex==2) { //Rm stat //local
        SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
        SoNode* sg = viewer->getSceneGraph();
        if(sg) {
          SoNode* node = (SoNode*)SoTools_searchNode(sg,STATIC_SCENE);
          if(node && node->isOfType(SoGroup::getClassTypeId())) {
            ((SoGroup*)node)->removeAllChildren();
          }
        }
        aEventCB->setHandled();

      } if(menu->fIndex==3) { //Rm dyn //local
        SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
        SoNode* sg = viewer->getSceneGraph();
        if(sg) {
          SoNode* node = (SoNode*)SoTools_searchNode(sg,DYNAMIC_SCENE);
          if(node && node->isOfType(SoGroup::getClassTypeId())) {
            ((SoGroup*)node)->removeAllChildren();
          }
        }
        aEventCB->setHandled();

      } if(menu->fIndex==4) { //ATLAS //local
        SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
        SoNode* sg = viewer->getSceneGraph();
        if(sg) {
          SoNode* node = (SoNode*)SoTools_searchNode(sg,DYNAMIC_SCENE);
          if(node && node->isOfType(SoGroup::getClassTypeId())) {
            ((SoGroup*)node)->removeAllChildren(); //clear.
            SoCurl* soCurl = new SoCurl;
            soCurl->name.setValue
     ("http://OpenScientist.lal.in2p3.fr/download/exp/ATLAS/iv/cosmics.iv");
            ((SoGroup*)node)->addChild(soCurl);
          }
        }
        aEventCB->setHandled();

      } if(menu->fIndex==5) { //LHCb //local
        SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
        SoNode* sg = viewer->getSceneGraph();
        if(sg) {
          SoNode* node = (SoNode*)SoTools_searchNode(sg,DYNAMIC_SCENE);
          if(node && node->isOfType(SoGroup::getClassTypeId())) {
            ((SoGroup*)node)->removeAllChildren(); //clear

            if(SoType::fromName("SoPolyhedron")==SoType::badType()){
              std::cout << "SoSDL_web :"
                       << " can't load LHCb file because HEPVis is not loaded."
                       << std::endl;
            } else {
              SoCurl* soCurl = new SoCurl;
              soCurl->name.setValue
  ("http://OpenScientist.lal.in2p3.fr/download/exp/LHCb/hiv/LHCb_prplot1.hiv");
              ((SoGroup*)node)->addChild(soCurl);
            }
          }
        }
        aEventCB->setHandled();
#endif //SOSDL_WEB
      }
    }

  } else if(SO_KEY_PRESS_EVENT(event,RIGHT_ARROW)) {
    Menu* menu = (Menu*)aTag;    
    SoSwitch* sw = menu->fSwitch;    
    if(sw->whichChild.getValue()==1) {
      next_item(*menu);
      aEventCB->setHandled();
    }

  } else if(SO_KEY_PRESS_EVENT(event,LEFT_ARROW)) {
    Menu* menu = (Menu*)aTag;    
    SoSwitch* sw = menu->fSwitch;    
    if(sw->whichChild.getValue()==1) {
      prev_item(*menu);
      aEventCB->setHandled();
    }

  }
}

#include <Inventor/SDL/SoJoyAxisEvent.h>

static void menu_joyAxisCB(void* aTag,SoEventCallback* aEventCB){
  if(aEventCB->isHandled()) return;
  const SoJoyAxisEvent* event = (SoJoyAxisEvent*)aEventCB->getEvent();
  Menu* menu = (Menu*)aTag;    
  SoSwitch* sw = menu->fSwitch;    
  if(sw->whichChild.getValue()==0) return;
  if(event->getState()==SoJoyAxisEvent::RIGHT) {
    Menu* menu = (Menu*)aTag;    
    SoSwitch* sw = menu->fSwitch;    
    if(sw->whichChild.getValue()==1) {
      next_item(*menu);
      aEventCB->setHandled();
    }
  } else if(event->getState()==SoJoyAxisEvent::LEFT) {
    Menu* menu = (Menu*)aTag;    
    SoSwitch* sw = menu->fSwitch;    
    if(sw->whichChild.getValue()==1) {
      prev_item(*menu);
      aEventCB->setHandled();
    }
  }
}

#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoText3.h>

static void add_button(SoSeparator& aButtons,const SbString& aLabel,float a_w,float a_h,SbBool a_highlight = FALSE) {
  SoSeparator* button = new SoSeparator;
  aButtons.addChild(button);

  //CHILD_TRANS 0
 {SoTranslation* trans = new SoTranslation;
  int n = aButtons.getNumChildren();
  //trans->translation.setValue(SbVec3f(n*a_w,0,0));
  if(n==1) {
    trans->translation.setValue(SbVec3f(a_w,0,0));
  } else {
    SoSeparator* pbutton = (SoSeparator*)aButtons.getChild(n-2);
    SoTranslation* ptrans = (SoTranslation*)pbutton->getChild(CHILD_TRANS);
    float px = ptrans->translation.getValue()[0];

    SoCube* pback = (SoCube*)pbutton->getChild(CHILD_BACK);
    float pw = pback->width.getValue();

    trans->translation.setValue(SbVec3f(px+pw*0.5F+a_w*0.5F,0,0));
  }
  button->addChild(trans);}

 {SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(0,0,0));
  button->addChild(material);

  //CHILD_BACK 2
  SoCube* soBack = new SoCube;
  soBack->width.setValue(a_w);
  soBack->height.setValue(a_h);
  soBack->depth.setValue(0.1F);
  button->addChild(soBack);}

  //CHILD_MAT 3
 {SoMaterial* material = new SoMaterial;
   material->diffuseColor.setValue(a_highlight?hg_color():bg_color());
  button->addChild(material);

  SoCube* soHBack = new SoCube;
  soHBack->width.setValue(a_w*0.95F);
  soHBack->height.setValue(a_h*0.95F);
  soHBack->depth.setValue(0.11F);
  button->addChild(soHBack);}

  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(0,0,0));
  button->addChild(material);

 {SoTranslation* trans = new SoTranslation;
  trans->translation.setValue(SbVec3f(0,-a_h*0.5F*0.8F,0.2));
  button->addChild(trans);}

  SoScale* scale = new SoScale;
  float fac = a_h;
  //float fac = 1;
  scale->scaleFactor.setValue(SbVec3f(fac,fac,1));
  button->addChild(scale);

  SoText3* soText = new SoText3;
  //SoText2* soText = new SoText2;
  soText->parts.setValue(SoText3::ALL);
  soText->string.setValue(aLabel);
  soText->justification.setValue(SoText3::CENTER);
  button->addChild(soText);

}
