#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoCube.h>

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/events/SoLocation2Event.h>

class Shoot : public SoSeparator {
public:
  Shoot():fSwitch(0),fTranslation(0){}
  SoSwitch* fSwitch;
  SoOrthographicCamera* fCamera;
  SoTranslation* fTranslation;
  unsigned int f_ww;
  unsigned int f_wh;
#ifdef SDL_WILD
  unsigned int f_rows;
  unsigned int f_col;
  unsigned int f_row;
  unsigned int f_WILD_ww;
  unsigned int f_WILD_wh;
  float f_WILD_aspect;
#else
  float f_aspect;
#endif
  //in local window coordinate (y=0 = top).
  int f_shoot_x;
  int f_shoot_y;
};

static void shoot_keyboard_CB(void* aTag,SoEventCallback* aEventCB){
  if(aEventCB->isHandled()) return;
  const SoKeyboardEvent* event = (SoKeyboardEvent*)aEventCB->getEvent();
  if(SO_KEY_PRESS_EVENT(event,R)) {
    //printf("debug : R typed\n");
    Shoot* shoot = (Shoot*)aTag;    
    SoSwitch* sw = shoot->fSwitch;
    sw->whichChild.setValue(sw->whichChild.getValue()?0:1);
    aEventCB->setHandled();

  } else if(SO_KEY_PRESS_EVENT(event,RETURN)) {
    //printf("debug : RETURN typed\n");
    Shoot* shoot = (Shoot*)aTag;    
    SoSwitch* sw = shoot->fSwitch;
    if(sw->whichChild.getValue()==1) {
      SoSDL::emulateLeftButtonDown(SbVec2s(shoot->f_shoot_x,shoot->f_shoot_y));
      aEventCB->setHandled();
    }

  } else if(SO_KEY_RELEASE_EVENT(event,RETURN)) {
    Shoot* shoot = (Shoot*)aTag;    
    SoSwitch* sw = shoot->fSwitch;
    if(sw->whichChild.getValue()==1) {
      SoSDL::emulateLeftButtonUp(SbVec2s(shoot->f_shoot_x,shoot->f_shoot_y));
      aEventCB->setHandled();
    }
  }
}

static void cvt_WILD_to_local(
 unsigned int a_col,unsigned int a_row
,unsigned int a_ww,unsigned int a_wh
,int& a_x,int& a_y) {
  unsigned int left = a_col * a_ww;
  unsigned int bottom = a_row * a_wh;
  a_x = a_x-left;
  a_y = a_y-bottom;
}

static void cvt_local_to_WILD(
 unsigned int a_col,unsigned int a_row
,unsigned int a_ww,unsigned int a_wh
,int& a_x,int& a_y) {
  unsigned int left = a_col * a_ww;
  unsigned int bottom = a_row * a_wh;
  a_x = a_x+left;
  a_y = a_y+bottom;
}

static void shoot_pos(Shoot& aShoot,SbVec3f& aPos){
  // aShoot.f_shoot_[x,y] is in local window coordinate.
  int x = aShoot.f_shoot_x;
  int y = aShoot.f_wh-aShoot.f_shoot_y; //y 0 = bottom
#ifdef SDL_WILD
  cvt_local_to_WILD(aShoot.f_col,aShoot.f_row,aShoot.f_ww,aShoot.f_wh,x,y);
  float f_x = float(x)/float(aShoot.f_WILD_ww);
  float f_y = float(y)/float(aShoot.f_WILD_wh);
  float h = aShoot.fCamera->height.getValue() * aShoot.f_rows;
  float w = aShoot.f_WILD_aspect * h;
#else
  float f_x = float(x)/float(aShoot.f_ww);
  float f_y = float(y)/float(aShoot.f_wh);
  float h = aShoot.fCamera->height.getValue();
  float w = aShoot.f_aspect * h;
#endif
  float tx = w*(f_x-0.5F);
  float ty = h*(f_y-0.5F);
  aPos.setValue(tx,ty,0);
}

static void shoot_location2_CB(void* aTag,SoEventCallback* aEventCB){
  if(aEventCB->isHandled()) return;
  Shoot* shoot = (Shoot*)aTag;    
  SoSwitch* sw = shoot->fSwitch;
  if(sw->whichChild.getValue()==1) {
    const SoLocation2Event* event = (SoLocation2Event*)aEventCB->getEvent();
    const SbVec2s& pos = event->getPosition(); //in local window coord.
    //printf("debug : pos : 000 : %d %d\n",pos[0],pos[1]);

    SDL_Surface* screen = SoSDL::screen();
    //printf("debug : surf %d %d\n",screen->w,screen->h);
    unsigned int ww = screen->w;
    unsigned int wh = screen->h;

    shoot->f_shoot_x = pos[0];
    shoot->f_shoot_y = wh-pos[1];

   {SoTranslation* trans = shoot->fTranslation;
    SbVec3f pos;
    shoot_pos(*shoot,pos);
    trans->translation.setValue(pos);}

    aEventCB->setHandled();
  }
}


