//  Mai/2007 : G.Barrand : creation.

#ifndef SoSDL_Viewer_h
#define SoSDL_Viewer_h 

#include <Inventor/SDL/SoSDL.h>

#include <Inventor/SoType.h>
#include <Inventor/SbVec2s.h>
#include <Inventor/SbVec3f.h>
#include <Inventor/SbRotation.h>

class SbVec3f;
class SbColor;
class SbViewportRegion;

class SoSceneManager;
class SoNode;
class SoGLRenderAction;
class SoNodeSensor;
class SoSensor;
class SoEvent;
class SoGroup;
class SoCamera;
class SoDirectionalLight;

typedef void(*SoSDL_UnlockCB)(void*);

class SoSDL_Viewer {
public:
  SoSDL_Viewer(SoSDL_Widget* = 0);
  virtual ~SoSDL_Viewer();
  
  virtual SbBool setSceneGraph(SoNode*);
  virtual SoNode* getSceneGraph() const;
  virtual void setViewportRegion(const SbViewportRegion&);
  virtual const SbViewportRegion& getViewportRegion() const;
  virtual void setBackgroundColor(const SbColor&);
  virtual const SbColor& getBackgroundColor() const;
  virtual void render();
  virtual void setSize(const SbVec2s&);
  virtual void viewAll();
  virtual void setViewing(SbBool);
  virtual void show();
  virtual void setGLRenderAction(SoGLRenderAction*);
  virtual SoGLRenderAction* getGLRenderAction() const;
  virtual SoSceneManager* getSceneManager() const;
  virtual void setTitle(const char*);
  virtual void setCamera(SoCamera*);
  virtual SoCamera* getCamera() const;
  virtual void setCameraType(SoType);
  virtual SoType getCameraType(void) const;
  virtual void setHeadlight(SbBool);
  virtual void setAutoClipping(SbBool);
  virtual void setDecoration(SbBool);
  
  virtual SbBool processEvent(const SDL_Event&);

  void rotateCameraPositionAroundX(float);
  void rotateCameraPositionAroundY(float);
  void rotateCameraAroundUp(float);
  void translateCameraAlongDir(float);
  void translateCameraAlongSide(float);
  void translateCameraAlongUp(float);

  float getDefaultDeltaRotation() const;
  float getDefaultDeltaTranslation() const;
  void setDefaultDeltaRotation(float angle_radian);
  void setDefaultDeltaTranslation(float);

  float getFloor() const;
  void setFloor(float);
  void checkFloor(SbBool);

  void updateMotion(SoSDL::Move);

  void saveCamera();
  void restoreCamera();

private:
  void setCameraPosition(const SbVec3f&);
private:
  SoSceneManager* fSceneManager;
  SoGroup* fSceneGraph;
  SoNode* fUserSceneGraph;
  SoCamera* fCamera;
  SoDirectionalLight* fHeadLight;
  SoType fCameraType;
  float fDefaultDeltaRotation;
  float fDefaultDeltaTranslation;  
  SbBool fCheckFloor;
  float fFloor;
private:
  SbBool fJoy_state;
  //
  SbVec3f fSave_camera_position;
  SbRotation fSave_camera_orientation;
  //float fSave_camera_nearDistance;
  //float fSave_camera_farDistance;
  //float fSave_camera_focalDistance;
};

#endif
