//  Mai/2007 : G.Barrand : creation.

// The scene is a grey floor with a red cube over it.
// At startup, the camera is slightly on top of the floor
// looking at the cube.

#include <Inventor/SDL/SoSDL.h>
#include <Inventor/SDL/viewers/SoSDL_Viewer.h>

#include <Inventor/SDL/app.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDirectionalLight.h>

#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoText3.h>

#ifdef SDL_WILD
#include <Lib/InternetSocket.h>
#include "camera.h"
#include "net.h"
#else
#include <math.h>
#ifndef M_PI
#define M_PI       3.1415926535897931160E0
#define M_PI_2     1.5707963267948965580E0  
#endif
#define FM_PI      ((float)M_PI)
#define FM_PI_2    ((float)M_PI_2)
#include <Inventor/nodes/SoPerspectiveCamera.h>
#endif

#include "../applications/SoTools.h"

#define HAS_MENU

#ifdef HAS_MENU
#include "../applications/menu.h"
#endif

//
#define HAS_SHOOT

#ifdef HAS_SHOOT
#include "../applications/shoot.h"
#endif

#define HAS_SELECTION

#ifdef HAS_SELECTION
#include "../applications/selection.h"
#endif

//Windows, Darwin : "main" becomes "SDL_main" through a cpp macro in SDL.h.

#include <Slash/Tools/args.h>
#include <Slash/Tools/system.h>

#include <iostream>
#include <cstdlib>

int main(int argc,char* argv[]) {

  bool is_mac_app = false;
#ifdef __APPLE__
  //NOTE : we can't test for argv[1] with -psn_ since
  //       SDLMain/main strips the argv[1] argument.
  std::string contents_dir;
  if(argc>0) {
    std::string prog_dir = Slash::dir_name(argv[0]);
    if(Slash::base_name(prog_dir)=="MacOS") {
      is_mac_app = true;
      contents_dir = Slash::dir_name(prog_dir);
    }
  }
#endif

  ///////////////////////////////////////////////
  // Init :
  ///////////////////////////////////////////////
  Slash::args args(argc,argv);

  int verbose = 0;
  if(is_mac_app) { 
    if(!Slash::get_env<int>("SOSDL_VERBOSE",verbose)) verbose = 0;
  } else {
    if(args.is_arg("-verbose")) verbose = 1;
  }

  bool full_screen = true; //default
  if(is_mac_app) { 
    if(::getenv("SOSDL_NOT_FULL_SCREEN")) full_screen = false;
  } else {
    if(args.is_arg("-not_full_screen")) full_screen = false;
  }

  SoSDL_Widget* topWidget = 0;
  if(full_screen) {
    topWidget = SoSDL::init(0,0,SDL_FULLSCREEN);
  } else {
    float aspect = 1680.0F/1050.0F;
    unsigned int ww = 800;
    unsigned int wh = (((float)ww)/aspect); //inside window height
    topWidget = SoSDL::init(ww,wh,0);
  }
  if(!topWidget) return EXIT_FAILURE;

  SDL_Surface* screen = SoSDL::screen();
  if(!screen) return EXIT_FAILURE;

  //printf("debug : surf %d %d\n",screen->w,screen->h);
  unsigned int ww = screen->w;
  unsigned int wh = screen->h;
  float aspect = ((float)ww)/((float)wh);

#ifdef SDL_WILD
#include "../applications/WILD_start.icc"
#else
  SDL_ShowCursor(1);
#endif

#ifdef HAS_SHOOT
#ifdef SDL_WILD
  //arrange to have the mire not on a screen separator.
  float shoot_fact_x = (2*int(screen_cols/2)==screen_cols)?0.55F:0.5F;
  float shoot_fact_y = (2*int(screen_rows/2)==screen_rows)?0.55F:0.5F;
  int shoot_x = WILD_ww*shoot_fact_x;
  int shoot_y = WILD_wh*shoot_fact_y;
  cvt_WILD_to_local(screen_col,screen_row,ww,wh,shoot_x,shoot_y);
#else
  float shoot_fact_x = 0.5F;
  float shoot_fact_y = 0.5F;
  int shoot_x = ww*shoot_fact_x;
  int shoot_y = wh*shoot_fact_y;
#endif
  Shoot* shoot = new Shoot;
  shoot->f_shoot_x = shoot_x;
  shoot->f_shoot_y = wh-shoot_y;  // 0 = top
#endif

  SoSeparator* root = new SoSeparator;
  root->ref();

#ifdef HAS_MENU
#include "../applications/menu.ic"
#endif

#ifdef HAS_SHOOT
#include "../applications/shoot.ic"
#endif

#ifdef HAS_SELECTION
  MySelection* sceneGraph = new MySelection;
  sceneGraph->addSelectionCallback(selectionCB,(void*)sceneGraph);
  sceneGraph->addDeselectionCallback(deselectionCB,(void*)sceneGraph);
  sceneGraph->setPickRoot(sceneGraph);
  sceneGraph->policy = SoSelection::TOGGLE;
  sceneGraph->policy = SoSelection::SINGLE;
  sceneGraph->policy = SoSelection::SHIFT;
#else
  SoSeparator* sceneGraph = new SoSeparator;
#endif
  root->addChild(sceneGraph);

  JoyTag joyTag;
  add_CBs(sceneGraph,&joyTag);

  ///////////////////////////////////////////////
  // Camera (us) :
  ///////////////////////////////////////////////
  // Head light :
 {SoDirectionalLight* soLight = new SoDirectionalLight;
  soLight->color.setValue(1,1,1);
  // Same as Coin<Gui>/So<Gui>Viewer :
  soLight->direction.setValue(SbVec3f(1,-1,-10));
  sceneGraph->addChild(soLight);}

  float d = 100;
  float eye = 7;
  float near = d*0.01F;
  float far = 100*d;
#ifdef SDL_WILD
  SoCamera* soCamera = create_camera(screen_cols,screen_rows,
		  		     screen_col,screen_row,
                                     aspect,2,
                                     near);
#else
  SoPerspectiveCamera* soCamera = new SoPerspectiveCamera;
  soCamera->nearDistance.setValue(near);
#endif
  soCamera->position.setValue(0,eye,d);
  soCamera->focalDistance.setValue(d);
  soCamera->farDistance.setValue(far);
  sceneGraph->addChild(soCamera);

  // Light at ceil :
 {SoPointLight* soLight = new SoPointLight;
  soLight->location.setValue(0,100,0); //y = 100.
  sceneGraph->addChild(soLight);}

  SoLightModel* soLightModel = new SoLightModel;
  soLightModel->model.setValue(SoLightModel::PHONG);
  sceneGraph->addChild(soLightModel);

  ///////////////////////////////////////////////
  // Floor :
  ///////////////////////////////////////////////
 {SoSeparator* floor = new SoSeparator;
  sceneGraph->addChild(floor);
  
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(1,1,1));
  floor->addChild(material);

  SoCube* soCube = new SoCube;
  soCube->width.setValue(50);
  soCube->height.setValue(0.1F);
  soCube->depth.setValue(50);
  floor->addChild(soCube);}

  ///////////////////////////////////////////////
  // Scene :
  ///////////////////////////////////////////////
 {SoSeparator* scene = new SoSeparator;
  sceneGraph->addChild(scene);

 {SoSeparator* sep = new SoSeparator;
  scene->addChild(sep);
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(0,1,0));
  sep->addChild(material);
  SoTransform* transform = new SoTransform;
  transform->translation.setValue(SbVec3f(-10,10,0));
  sep->addChild(transform);
  SoCube* soCube = new SoCube;
  soCube->width.setValue(10.0F);
  soCube->height.setValue(10.0F);
  soCube->depth.setValue(10.0F);
  sep->addChild(soCube);}

 {SoSeparator* sep = new SoSeparator;
  scene->addChild(sep);
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(0,0,1));
  sep->addChild(material);
  SoTransform* transform = new SoTransform;
  transform->translation.setValue(SbVec3f(10,10,0));
  sep->addChild(transform);
  SoSphere* soSphe = new SoSphere;
  soSphe->radius.setValue(5.0F);
  sep->addChild(soSphe);}

 {SoSeparator* sep = new SoSeparator;
  scene->addChild(sep);
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(1,1,0));
  sep->addChild(material);
  SoTransform* transform = new SoTransform;
  transform->translation.setValue(SbVec3f(0,10,0));
  sep->addChild(transform);
  SoFont* font = new SoFont;
  //font->name.setValue("Helvetica");
  font->size.setValue(5.0F);
  sep->addChild(font);    
  SoText3* soText = new SoText3;
  soText->parts.setValue(SoText3::ALL);
  soText->string.setValue("Text");
  soText->justification.setValue(SoText3::CENTER);
  sep->addChild(soText);}

  } //end scene

  // Viewer :
  SoSDL_Viewer* viewer = new SoSDL_Viewer(topWidget);
#ifdef HAS_SELECTION
  viewer->setGLRenderAction(new SoBoxHighlightRenderAction());
  //viewer->setGLRenderAction(new SoLineHighlightRenderAction());
#endif
  viewer->setSceneGraph(root);
  viewer->setCamera(soCamera);
  viewer->setHeadlight(FALSE);
  viewer->setDefaultDeltaTranslation(1);

  //SoSDL :
  viewer->saveCamera();

  viewer->show();
    
  SoSDL::show(topWidget);

  SoSDL::mainLoop();
    
  root->unref();

  delete viewer;

  SoSDL::clear();  
	
#ifdef SDL_WILD
  delete socket;
#endif

  if(verbose) std::cout << "SoSDL viewer : exiting..." << std::endl;

  return EXIT_SUCCESS;
}
